import numpy as np


class LoadVesuvioBackParameters:
    def __init__(self, ipFilesPath):
        self.ipfile = ipFilesPath / "ip2019.par"

    runs = "43066-43076"  # 77K         # The numbers of the runs to be analysed
    empty_runs = (
        "41876-41923"  # 77K         # The numbers of the empty runs to be subtracted
    )
    mode = "DoubleDifference"

    subEmptyFromRaw = True  # Flag to control wether empty ws gets subtracted from raw
    scaleEmpty = 1  # None or scaling factor
    scaleRaw = 1


class LoadVesuvioFrontParameters:
    def __init__(self, ipFilesPath):
        self.ipfile = ipFilesPath / "ip2018_3.par"

    runs = "43066-43076"  # 100K        # The numbers of the runs to be analysed
    empty_runs = (
        "43868-43911"  # 100K        # The numbers of the empty runs to be subtracted
    )
    mode = "SingleDifference"

    subEmptyFromRaw = False  # Flag to control wether empty ws gets subtracted from raw
    scaleEmpty = 1  # None or scaling factor
    scaleRaw = 1


class GeneralInitialConditions:
    """Used to define initial conditions shared by both Back and Forward scattering"""

    # Sample slab parameters
    vertical_width, horizontal_width, thickness = 0.1, 0.1, 0.001  # Expressed in meters


class BackwardInitialConditions(GeneralInitialConditions):
    def __init__(self, ipFilesPath):
        self.InstrParsPath = ipFilesPath / "ip2018_3.par"

    HToMassIdxRatio = 19.0620008206  # Set to zero or None when H is not present
    massIdx = 0

    # Masses, instrument parameters and initial fitting parameters
    masses = np.array([12, 16, 27])
    # noOfMasses = len(masses)

    # Intensities, NCP widths, NCP centers
    initPars = np.array([1, 12, 0.0, 1, 12, 0.0, 1, 12.5, 0.0])
    bounds = np.array(
        [
            [0, np.nan],
            [8, 16],
            [-3, 1],
            [0, np.nan],
            [8, 16],
            [-3, 1],
            [0, np.nan],
            [11, 14],
            [-3, 1],
        ]
    )
    constraints = ()

    noOfMSIterations = 3  # 4
    firstSpec = 3  # 3
    lastSpec = 134  # 134

    maskedSpecAllNo = np.array([18, 34, 42, 43, 59, 60, 62, 118, 119, 133])

    # Boolean Flags to control script
    MSCorrectionFlag = True
    GammaCorrectionFlag = False

    # # Parameters of workspaces in input_ws
    tofBinning = "275.,1.,420"  # Binning of ToF spectra
    maskTOFRange = None

    transmission_guess = 0.8537  # Experimental value from VesuvioTransmission
    multiple_scattering_order = 2
    number_of_events = 1.0e5

    # Original data uses histogram data instead of point data
    runHistData = True
    normVoigt = False


class ForwardInitialConditions(GeneralInitialConditions):
    def __init__(self, ipFilesPath):
        self.InstrParsPath = ipFilesPath / "ip2018_3.par"

    masses = np.array([1.0079, 12, 16, 27])

    # Intensities, NCP widths, NCP centers
    initPars = np.array([1, 4.7, 0, 1, 12.71, 0.0, 1, 8.76, 0.0, 1, 13.897, 0.0])
    bounds = np.array(
        [
            [0, np.nan],
            [3, 6],
            [-3, 1],
            [0, np.nan],
            [12.71, 12.71],
            [-3, 1],
            [0, np.nan],
            [8.76, 8.76],
            [-3, 1],
            [0, np.nan],
            [13.897, 13.897],
            [-3, 1],
        ]
    )
    constraints = ()

    noOfMSIterations = 0  # 4
    firstSpec = 144  # 144
    lastSpec = 182  # 182

    # Boolean Flags to control script
    MSCorrectionFlag = True
    GammaCorrectionFlag = True

    maskedSpecAllNo = np.array([173, 174, 179])

    tofBinning = "110,1,430"  # Binning of ToF spectra
    maskTOFRange = None

    transmission_guess = 0.8537  # Experimental value from VesuvioTransmission
    multiple_scattering_order = 2
    number_of_events = 1.0e5

    # Original data uses histogram data instead of point data
    runHistData = True
    normVoigt = False


class YSpaceFitInitialConditions:
    showPlots = True
    symmetrisationFlag = True
    rebinParametersForYSpaceFit = "-25, 0.5, 25"  # Needs to be symetric
    fitModel = "Gaussian3D"  # Options: 'SINGLE_GAUSSIAN', 'GC_C4', 'GC_C6', 'GC_C4_C6', 'DOUBLE_WELL', 'ANSIO_GAUSSIAN', 'Gaussian3D'
    runMinos = True
    globalFit = True  # Performs global fit with Minuit by default
    nGlobalFitGroups = 4  # Number or string "ALL"
    maskTypeProcedure = "NAN"  # Options: 'NCP', 'NAN', None


class UserScriptControls:
    runRoutine = True

    # Choose main procedure to run
    procedure = "FORWARD"  # Options: None, "BACKWARD", "FORWARD", "JOINT"

    # Choose on which ws to perform the fit in y space
    fitInYSpace = "FORWARD"  # Options: None, "BACKWARD", "FORWARD", "JOINT"


class BootstrapInitialConditions:
    runBootstrap = False



import time
from pathlib import Path
import importlib
import sys
from mvesuvio.vesuvio_analysis.run_script import runScript
from mvesuvio.scripts import handle_config


def run(yes_to_all=False):
    scriptName = handle_config.read_config_var("caching.experiment")
    experimentsPath = (
        Path(handle_config.read_config_var("caching.location"))
        / "experiments"
        / scriptName
    )  # Path to the repository
    inputs_path = experimentsPath / "analysis_inputs.py"
    ipFilesPath = Path(handle_config.read_config_var("caching.ipfolder"))
    ai = import_from_path(inputs_path, "analysis_inputs")

    start_time = time.time()

    wsBackIC = ai.LoadVesuvioBackParameters(ipFilesPath)
    wsFrontIC = ai.LoadVesuvioFrontParameters(ipFilesPath)
    bckwdIC = ai.BackwardInitialConditions(ipFilesPath)
    fwdIC = ai.ForwardInitialConditions
    yFitIC = ai.YSpaceFitInitialConditions
    bootIC = ai.BootstrapInitialConditions
    userCtr = ai.UserScriptControls

    runScript(
        userCtr,
        scriptName,
        wsBackIC,
        wsFrontIC,
        bckwdIC,
        fwdIC,
        yFitIC,
        bootIC,
        yes_to_all,
    )

    end_time = time.time()
    print("\nRunning time: ", end_time - start_time, " seconds")


def import_from_path(path, name):
    spec = importlib.util.spec_from_file_location(name, path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[name] = module
    spec.loader.exec_module(module)
    return module


if __name__ == "__main__":
    run()



from mvesuvio.vesuvio_analysis.run_script import runScript
import unittest
import numpy as np
import numpy.testing as nptest
from pathlib import Path
from mvesuvio.system_tests.tests_IC import (
    scriptName,
    wsBackIC,
    wsFrontIC,
    bckwdIC,
    fwdIC,
    yFitIC,
)


class BootstrapInitialConditions:
    runBootstrap = True

    procedure = "JOINT"
    fitInYSpace = None

    bootstrapType = "JACKKNIFE"
    nSamples = 3  # Overwritten by running Jackknife
    skipMSIterations = False
    runningTest = True
    userConfirmation = False


class UserScriptControls:
    runRoutine = False
    procedure = "FORWARD"
    fitInYSpace = None
    # bootstrap = "JOINT"


class TestJointBootstrap(unittest.TestCase):
    @classmethod
    def _run_analysis(cls):
        np.random.seed(3)  # Set seed so that tests match everytime

        bootIC = BootstrapInitialConditions
        userCtr = UserScriptControls

        bootRes, noneRes = runScript(
            userCtr, scriptName, wsBackIC, wsFrontIC, bckwdIC, fwdIC, yFitIC, bootIC, True
        )

        cls._jackBackSamples = bootRes["bckwdScat"].bootSamples
        cls._jackFrontSamples = bootRes["fwdScat"].bootSamples

        # jackJointResults = bootRes

        # jackSamples = []
        # for jackRes in jackJointResults:
        #     jackSamples.append(jackRes.bootSamples)

        # jackBackSamples, jackFrontSamples = jackSamples

    @classmethod
    def _load_benchmark_results(cls):
        testPath = Path(__file__).absolute().parent
        cls._oriJointBack = np.load(str(testPath / "stored_joint_jack_back.npz"))[
            "boot_samples"
        ]
        cls._oriJointFront = np.load(str(testPath / "stored_joint_jack_front.npz"))[
            "boot_samples"
        ]

    @classmethod
    def setUpClass(cls):
        cls._run_analysis()
        cls._load_benchmark_results()

    def xtestBack(self):
        nptest.assert_array_almost_equal(self._jackBackSamples, self._oriJointBack)

    def xtestFront(self):
        nptest.assert_array_almost_equal(self._jackFrontSamples, self._oriJointFront)


if __name__ == "__main__":
    unittest.main()



import os
from shutil import copyfile, copytree, ignore_patterns


def __parse_config_env_var():
    env_var = os.getenv("VESUVIOPROPERTIES")
    if env_var:
        config_path, config_file = os.path.split(env_var)
    else:
        config_path = os.path.join(os.path.expanduser("~"), ".mvesuvio")
        config_file = "vesuvio.user.properties"
    return config_path, config_file


### PATH CONSTANTS ###
VESUVIO_CONFIG_PATH, VESUVIO_CONFIG_FILE = __parse_config_env_var()
VESUVIO_INPUTS_FILE = "analysis_inputs.py"
VESUVIO_PACKAGE_PATH = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
MANTID_CONFIG_FILE = "Mantid.user.properties"
VESUVIO_IPFOLDER_PATH = os.path.join(VESUVIO_CONFIG_PATH, "ip_files")
######################


def __read_config(config_file_path, throw_on_not_found=True):
    lines = ""
    try:
        with open(config_file_path, "r") as file:
            lines = file.readlines()
    except IOError:
        if throw_on_not_found:
            raise RuntimeError(
                f"Could not read from vesuvio config file: {config_file_path}"
            )
    return lines


def set_config_vars(var_dict):
    file_path = os.path.join(VESUVIO_CONFIG_PATH, VESUVIO_CONFIG_FILE)
    lines = __read_config(file_path)

    updated_lines = []
    for line in lines:
        match = False
        for var in var_dict:
            if line.startswith(var):
                new_line = f"{var}={var_dict[var]}"
                updated_lines.append(f"{new_line}\n")
                match = True
                print(f"Setting: {new_line}")
                break

        if not match:
            updated_lines.append(line)

    with open(file_path, "w") as file:
        file.writelines(updated_lines)


def read_config_var(var, throw_on_not_found=True):
    file_path = f"{VESUVIO_CONFIG_PATH}{os.path.sep}{VESUVIO_CONFIG_FILE}"
    lines = __read_config(file_path, throw_on_not_found)

    result = ""
    for line in lines:
        if line.startswith(var):
            result = line.split("=", 2)[1].strip("\n")
            break
    if not result and throw_on_not_found:
        raise ValueError(f"{var} was not found in the vesuvio config")
    return result


def setup_config_dir(config_dir):
    success = __mk_dir("config", config_dir)
    if success:
        copyfile(
            os.path.join(VESUVIO_PACKAGE_PATH, "config", VESUVIO_CONFIG_FILE),
            os.path.join(config_dir, VESUVIO_CONFIG_FILE),
        )
        copyfile(
            os.path.join(VESUVIO_PACKAGE_PATH, "config", MANTID_CONFIG_FILE),
            os.path.join(config_dir, MANTID_CONFIG_FILE),
        )


def setup_expr_dir(cache_dir, experiment):
    expr_path = os.path.join(cache_dir, "experiments", experiment)
    success = __mk_dir("experiment", expr_path)
    if success:
        copyfile(
            os.path.join(VESUVIO_PACKAGE_PATH, "config", VESUVIO_INPUTS_FILE),
            input_file_path(cache_dir, experiment),
        )


def setup_default_ipfile_dir():
    if not os.path.isdir(VESUVIO_IPFOLDER_PATH):
        copytree(
            os.path.join(VESUVIO_PACKAGE_PATH, "config", "ip_files"),
            VESUVIO_IPFOLDER_PATH,
            ignore=ignore_patterns("__*"),
        )


def __mk_dir(type, path):
    success = False
    if not os.path.isdir(path):
        try:
            os.makedirs(path)
            success = True
        except:
            print(f"Unable to make {type} directory at location: {path}")
    return success


def config_set():
    if read_config_var("caching.location", False):
        return True
    else:
        return False


def input_file_path(cache_dir, experiment):
    return os.path.join(cache_dir, "experiments", experiment, VESUVIO_INPUTS_FILE)


def check_dir_exists(type, path):
    if not os.path.isdir(path):
        print(f"Directory of {type} could not be found at location: {path}")
        return False
    return True



from mvesuvio.vesuvio_analysis.run_script import runScript
import unittest
import numpy as np
import numpy.testing as nptest
from pathlib import Path
from mvesuvio.scripts import handle_config
from mvesuvio.system_tests.analysis_inputs import (
    LoadVesuvioBackParameters,
    LoadVesuvioFrontParameters,
    BackwardInitialConditions,
    ForwardInitialConditions,
    YSpaceFitInitialConditions,
    BootstrapInitialConditions,
    UserScriptControls,
)


ipFilesPath = Path(handle_config.read_config_var("caching.ipfolder"))


class AnalysisRunner:
    _benchmarkResults = None
    _currentResults = None

    @classmethod
    def get_benchmark_result(cls):
        if not AnalysisRunner._benchmarkResults:
            cls._load_benchmark_results()
        return AnalysisRunner._benchmarkResults

    @classmethod
    def get_current_result(cls):
        if not AnalysisRunner._currentResults:
            cls._run()
        return AnalysisRunner._currentResults

    @classmethod
    def _run(cls):
        scattRes, yfitRes = runScript(
            UserScriptControls(),
            "test_expr",
            LoadVesuvioBackParameters(ipFilesPath),
            LoadVesuvioFrontParameters(ipFilesPath),
            BackwardInitialConditions(ipFilesPath),
            ForwardInitialConditions(ipFilesPath),
            YSpaceFitInitialConditions(),
            BootstrapInitialConditions(),
            True,
        )

        wsFinal, forwardScatteringResults = scattRes

        # Test the results
        np.set_printoptions(suppress=True, precision=8, linewidth=150)

        currentResults = forwardScatteringResults
        AnalysisRunner._currentResults = currentResults

    @classmethod
    def _load_benchmark_results(cls):
        testPath = Path(__file__).absolute().parent
        benchmarkResults = np.load(
            str(testPath / "stored_spec_144-182_iter_3_GC_MS.npz")
        )
        AnalysisRunner._benchmarkResults = benchmarkResults


class TestFitParameters(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.benchmarkResults = AnalysisRunner.get_benchmark_result()
        cls.currentResults = AnalysisRunner.get_current_result()

    def setUp(self):
        oriPars = self.benchmarkResults["all_spec_best_par_chi_nit"]
        self.orispec = oriPars[:, :, 0]
        self.orichi2 = oriPars[:, :, -2]
        self.orinit = oriPars[:, :, -1]
        self.orimainPars = oriPars[:, :, 1:-2]
        self.oriintensities = self.orimainPars[:, :, 0::3]
        self.oriwidths = self.orimainPars[:, :, 1::3]
        self.oricenters = self.orimainPars[:, :, 2::3]

        optPars = self.currentResults.all_spec_best_par_chi_nit
        self.optspec = optPars[:, :, 0]
        self.optchi2 = optPars[:, :, -2]
        self.optnit = optPars[:, :, -1]
        self.optmainPars = optPars[:, :, 1:-2]
        self.optintensities = self.optmainPars[:, :, 0::3]
        self.optwidths = self.optmainPars[:, :, 1::3]
        self.optcenters = self.optmainPars[:, :, 2::3]

    def test_chi2(self):
        nptest.assert_almost_equal(self.orichi2, self.optchi2, decimal=6)

    def test_nit(self):
        nptest.assert_almost_equal(self.orinit, self.optnit, decimal=-2)

    def test_intensities(self):
        nptest.assert_almost_equal(self.oriintensities, self.optintensities, decimal=2)

    def test_widths(self):
        nptest.assert_almost_equal(self.oriwidths, self.optwidths, decimal=2)

    def test_centers(self):
        nptest.assert_almost_equal(self.oricenters, self.optcenters, decimal=1)


class TestNcp(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.benchmarkResults = AnalysisRunner.get_benchmark_result()
        cls.currentResults = AnalysisRunner.get_current_result()

    def setUp(self):
        self.orincp = self.benchmarkResults["all_tot_ncp"][:, :, :-1]

        self.optncp = self.currentResults.all_tot_ncp

    def test_ncp(self):
        correctNansOri = np.where(
            (self.orincp == 0) & np.isnan(self.optncp), np.nan, self.orincp
        )
        nptest.assert_almost_equal(correctNansOri, self.optncp, decimal=4)


class TestMeanWidths(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.benchmarkResults = AnalysisRunner.get_benchmark_result()
        cls.currentResults = AnalysisRunner.get_current_result()

    def setUp(self):
        self.orimeanwidths = self.benchmarkResults["all_mean_widths"]
        self.optmeanwidths = self.currentResults.all_mean_widths

    def test_widths(self):
        nptest.assert_almost_equal(self.orimeanwidths, self.optmeanwidths, decimal=5)


class TestMeanIntensities(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.benchmarkResults = AnalysisRunner.get_benchmark_result()
        cls.currentResults = AnalysisRunner.get_current_result()

    def setUp(self):
        self.orimeanintensities = self.benchmarkResults["all_mean_intensities"]
        self.optmeanintensities = self.currentResults.all_mean_intensities

    def test_intensities(self):
        nptest.assert_almost_equal(self.orimeanintensities, self.optmeanintensities, decimal=6)


class TestFitWorkspaces(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.benchmarkResults = AnalysisRunner.get_benchmark_result()
        cls.currentResults = AnalysisRunner.get_current_result()

    def setUp(self):
        self.oriws = self.benchmarkResults["all_fit_workspaces"]
        self.optws = self.currentResults.all_fit_workspaces

    def test_ws(self):
        nptest.assert_almost_equal(self.oriws, self.optws, decimal=6)

if __name__ == "__main__":
    unittest.main()



from mantid.simpleapi import Load, LoadVesuvio, SaveNexus, DeleteWorkspace
from pathlib import Path
from mvesuvio.scripts import handle_config
from mantid.kernel import logger
import ntpath

experimentsPath = (
    Path(handle_config.read_config_var("caching.location")) / "experiments"
)


def completeICFromInputs(IC, scriptName, wsIC):
    """Assigns new methods to the initial conditions class from the inputs of that class"""

    assert (
        IC.lastSpec > IC.firstSpec
    ), "Last spectrum needs to be bigger than first spectrum"
    assert ((IC.lastSpec < 135) & (IC.firstSpec < 135)) | (
        (IC.lastSpec >= 135) & (IC.firstSpec >= 135)
    ), "First and last spec need to be both in Back or Front scattering."

    if IC.lastSpec <= 134:
        IC.modeRunning = "BACKWARD"
    elif IC.firstSpec >= 135:
        IC.modeRunning = "FORWARD"
    else:
        raise ValueError("Invalid first and last spectra input.")

    IC.name = scriptName + "_" + IC.modeRunning + "_"

    IC.masses = IC.masses.astype(float)
    IC.noOfMasses = len(IC.masses)

    IC.maskedSpecNo = IC.maskedSpecAllNo[
        (IC.maskedSpecAllNo >= IC.firstSpec) & (IC.maskedSpecAllNo <= IC.lastSpec)
    ]
    IC.maskedDetectorIdx = IC.maskedSpecNo - IC.firstSpec

    # Extract some attributes from wsIC
    IC.mode = wsIC.mode
    IC.subEmptyFromRaw = wsIC.subEmptyFromRaw
    IC.scaleEmpty = wsIC.scaleEmpty
    IC.scaleRaw = wsIC.scaleRaw

    # When attribute InstrParsPath is not present, set it equal to path from wsIC
    try:
        IC.InstrParsPath  # If present, leave it unaltered
    except AttributeError:
        IC.InstrParsPath = wsIC.ipfile

    # Sort out input and output paths
    rawPath, emptyPath = setInputWSForSample(wsIC, scriptName)

    IC.userWsRawPath = rawPath
    IC.userWsEmptyPath = emptyPath

    setOutputDirsForSample(IC, scriptName)

    # Do not run bootstrap sample, by default
    IC.runningSampleWS = False

    # Store script name
    IC.scriptName = scriptName

    # Default not running preliminary procedure to estimate HToMass0Ratio
    IC.runningPreliminary = False

    # Set directories for figures
    figSavePath = experimentsPath / scriptName / "figures"
    figSavePath.mkdir(exist_ok=True)
    IC.figSavePath = figSavePath

    # Create default of not running original version with histogram data
    try:
        IC.runHistData
    except AttributeError:
        IC.runHistData = False

    # Norm voigt except when comparing with tests
    try:
        IC.normVoigt
    except AttributeError:
        IC.normVoigt = True

    return


def setInputWSForSample(wsIC, sampleName):
    inputWSPath = experimentsPath / sampleName / "input_ws"
    inputWSPath.mkdir(parents=True, exist_ok=True)

    runningMode = getRunningMode(wsIC)

    rawWSName = sampleName + "_" + "raw" + "_" + runningMode + ".nxs"
    emptyWSName = sampleName + "_" + "empty" + "_" + runningMode + ".nxs"

    rawPath = inputWSPath / rawWSName
    emptyPath = inputWSPath / emptyWSName

    if not wsHistoryMatchesInputs(wsIC.runs, wsIC.mode, wsIC.ipfile, rawPath):
        saveWSFromLoadVesuvio(wsIC.runs, wsIC.mode, wsIC.ipfile, rawPath)

    if not wsHistoryMatchesInputs(wsIC.empty_runs, wsIC.mode, wsIC.ipfile, emptyPath):
        saveWSFromLoadVesuvio(wsIC.empty_runs, wsIC.mode, wsIC.ipfile, emptyPath)

    return rawPath, emptyPath


def getRunningMode(wsIC):
    if wsIC.__class__.__name__ == "LoadVesuvioBackParameters":
        runningMode = "backward"
    elif wsIC.__class__.__name__ == "LoadVesuvioFrontParameters":
        runningMode = "forward"
    else:
        raise ValueError(
            f"Input class for loading workspace not valid: {wsIC.__class__.__name__}"
        )
    return runningMode


def setOutputDirsForSample(IC, sampleName):
    outputPath = experimentsPath / sampleName / "output_files"
    outputPath.mkdir(parents=True, exist_ok=True)

    # Build Filename based on ic
    corr = ""
    if IC.GammaCorrectionFlag & (IC.noOfMSIterations > 0):
        corr += "_GC"
    if IC.MSCorrectionFlag & (IC.noOfMSIterations > 0):
        corr += "_MS"

    fileName = (
        f"spec_{IC.firstSpec}-{IC.lastSpec}_iter_{IC.noOfMSIterations}{corr}" + ".npz"
    )
    fileNameYSpace = fileName + "_ySpaceFit" + ".npz"

    IC.resultsSavePath = outputPath / fileName
    IC.ySpaceFitSavePath = outputPath / fileNameYSpace
    return


def wsHistoryMatchesInputs(runs, mode, ipfile, localPath):
    if not (localPath.is_file()):
        logger.notice("Cached workspace not found")
        return False
    local_ws = Load(Filename=str(localPath))
    ws_history = local_ws.getHistory()
    metadata = ws_history.getAlgorithmHistory(0)

    saved_runs = metadata.getPropertyValue("Filename")
    if saved_runs != runs:
        logger.notice(
            f"Filename in saved workspace did not match: {saved_runs} and {runs}"
        )
        return False

    saved_mode = metadata.getPropertyValue("Mode")
    if saved_mode != mode:
        logger.notice(f"Mode in saved workspace did not match: {saved_mode} and {mode}")
        return False

    saved_ipfile_name = ntpath.basename(metadata.getPropertyValue("InstrumentParFile"))
    if saved_ipfile_name != ipfile.name:
        logger.notice(
            f"IP files in saved workspace did not match: {saved_ipfile_name} and {ipfile.name}"
        )
        return False

    print("\nLocally saved workspace metadata matched with analysis inputs.\n")
    DeleteWorkspace(local_ws)
    return True


def saveWSFromLoadVesuvio(runs, mode, ipfile, localPath):
    if "backward" in localPath.name:
        spectra = "3-134"
    elif "forward" in localPath.name:
        spectra = "135-198"
    else:
        raise ValueError(f"Invalid name to save workspace: {localPath.name}")

    vesuvio_ws = LoadVesuvio(
        Filename=runs,
        SpectrumList=spectra,
        Mode=mode,
        InstrumentParFile=str(ipfile),
        OutputWorkspace=localPath.name,
        LoadLogFiles=False,
    )

    SaveNexus(vesuvio_ws, str(localPath.absolute()))
    print(f"Workspace saved locally at: {localPath.absolute()}")
    return


def completeBootIC(bootIC, bckwdIC, fwdIC, yFitIC):
    if not (bootIC.runBootstrap):
        return

    try:  # Assume it is not running a test if atribute is not found
        bootIC.runningTest
    except AttributeError:
        bootIC.runningTest = False

    setBootstrapDirs(bckwdIC, fwdIC, bootIC, yFitIC)
    return


def setBootstrapDirs(bckwdIC, fwdIC, bootIC, yFitIC):
    """Form bootstrap output data paths"""

    # Select script name and experiments path
    sampleName = bckwdIC.scriptName  # Name of sample currently running

    # Used to store running times required to estimate Bootstrap total run time.
    bootIC.runTimesPath = experimentsPath / sampleName / "running_times.txt"

    # Make bootstrap and jackknife data directories
    if bootIC.bootstrapType == "JACKKNIFE":
        bootPath = experimentsPath / sampleName / "jackknife_data"
    else:
        bootPath = experimentsPath / sampleName / "bootstrap_data"
    bootPath.mkdir(exist_ok=True)

    # Folders for skipped and unskipped MS
    if bootIC.skipMSIterations:
        dataPath = bootPath / "skip_MS_corrections"
    else:
        dataPath = bootPath / "with_MS_corrections"
    dataPath.mkdir(exist_ok=True)

    # Create text file for logs
    logFilePath = dataPath / "data_files_log.txt"
    if not (logFilePath.is_file()):
        with open(logFilePath, "w") as txtFile:
            txtFile.write(header_string())

    for IC in [bckwdIC, fwdIC]:  # Make save paths for .npz files
        bootName, bootNameYFit = genBootFilesName(IC, bootIC)

        IC.bootSavePath = (
            dataPath / bootName
        )  # works because modeRunning has same strings as procedure
        IC.bootYFitSavePath = dataPath / bootNameYFit

        IC.logFilePath = logFilePath
        IC.bootSavePathLog = logString(bootName, IC, yFitIC, bootIC, isYFit=False)
        IC.bootYFitSavePathLog = logString(
            bootNameYFit, IC, yFitIC, bootIC, isYFit=True
        )
    return


def genBootFilesName(IC, bootIC):
    """Generates save file name for either BACKWARD or FORWARD class"""

    nSamples = bootIC.nSamples
    if bootIC.bootstrapType == "JACKKNIFE":
        nSamples = 3 if bootIC.runningTest else noOfHistsFromTOFBinning(IC)

    # Build Filename based on ic
    corr = ""
    if IC.MSCorrectionFlag & (IC.noOfMSIterations > 0):
        corr += "_MS"
    if IC.GammaCorrectionFlag & (IC.noOfMSIterations > 0):
        corr += "_GC"

    fileName = f"spec_{IC.firstSpec}-{IC.lastSpec}_iter_{IC.noOfMSIterations}{corr}"
    bootName = fileName + f"_nsampl_{nSamples}" + ".npz"
    bootNameYFit = fileName + "_ySpaceFit" + f"_nsampl_{nSamples}" + ".npz"
    return bootName, bootNameYFit


def header_string():
    return """
    This file contains some information about each data file in the folder.
    ncp data file: boot type | procedure | tof binning | masked tof range.
    yspace fit data file: boot type | procedure | symmetrisation | rebin pars | fit model | mask type
    """


def logString(bootDataName, IC, yFitIC, bootIC, isYFit):
    if isYFit:
        log = (
            bootDataName
            + " : "
            + bootIC.bootstrapType
            + " | "
            + str(bootIC.fitInYSpace)
            + " | "
            + str(yFitIC.symmetrisationFlag)
            + " | "
            + yFitIC.rebinParametersForYSpaceFit
            + " | "
            + yFitIC.fitModel
            + " | "
            + str(yFitIC.maskTypeProcedure)
        )
    else:
        log = (
            bootDataName
            + " : "
            + bootIC.bootstrapType
            + " | "
            + str(bootIC.procedure)
            + " | "
            + IC.tofBinning
            + " | "
            + str(IC.maskTOFRange)
        )
    return log


def noOfHistsFromTOFBinning(IC):
    # Convert first to float and then to int because of decimal points
    start, spacing, end = [int(float(s)) for s in IC.tofBinning.split(",")]
    return int((end - start) / spacing) - 1  # To account for last column being ignored


def buildFinalWSName(scriptName: str, procedure: str, IC):
    # Format of corrected ws from last iteration
    name = scriptName + "_" + procedure + "_" + str(IC.noOfMSIterations)
    return name


def completeYFitIC(yFitIC, sampleName):
    # Set directories for figures

    figSavePath = experimentsPath / sampleName / "figures"
    figSavePath.mkdir(exist_ok=True)
    yFitIC.figSavePath = figSavePath
    return



import numpy as np
from mantid.simpleapi import *


def fitGlobalMantidFit(wsJoY, wsQ, wsRes, minimizer, yFitIC, wsFirstMass):
    """Uses original Mantid procedure to perform global fit on groups with 8 detectors"""

    replaceNansWithZeros(wsJoY)
    artificialErrorsInUnphysicalBins(wsJoY)
    createOneOverQWs(wsQ)
    mantidGlobalFitProcedure(wsGlobal, wsQInv, wsRes, minimizer, yFitIC, wsFirstMass)


def replaceNansWithZeros(ws):
    for j in range(ws.getNumberHistograms()):
        ws.dataY(j)[np.isnan(ws.dataY(j)[:])] = 0
        ws.dataE(j)[np.isnan(ws.dataE(j)[:])] = 0


def artificialErrorsInUnphysicalBins(wsJoY):
    wsGlobal = CloneWorkspace(
        InputWorkspace=wsJoY, OutputWorkspace=wsJoY.name() + "_Global"
    )
    for j in range(wsGlobal.getNumberHistograms()):
        wsGlobal.dataE(j)[wsGlobal.dataE(j)[:] == 0] = 0.1

    assert np.any(np.isnan(wsGlobal.extractE())) is False, (
        "Nan present in input workspace need to be replaced by " "zeros."
    )

    return wsGlobal


def createOneOverQWs(wsQ):
    wsInvQ = CloneWorkspace(InputWorkspace=wsQ, OutputWorkspace=wsQ.name() + "_Inverse")
    for j in range(wsInvQ.getNumberHistograms()):
        nonZeroFlag = wsInvQ.dataY(j)[:] != 0
        wsInvQ.dataY(j)[nonZeroFlag] = 1 / wsInvQ.dataY(j)[nonZeroFlag]

        ZeroIdxs = np.argwhere(wsInvQ.dataY(j)[:] == 0)  # Indxs of zero elements
        if ZeroIdxs.size != 0:  # When zeros are present
            wsInvQ.dataY(j)[ZeroIdxs[0] - 1] = 0  # Put a zero before the first zero

    return wsInvQ


def mantidGlobalFitProcedure(wsGlobal, wsQInv, wsRes, minimizer, yFitIC, wsFirstMass):
    """Original Implementation of Global Fit using Mantid"""

    if yFitIC.fitModel == "SINGLE_GAUSSIAN":
        convolution_template = """
        (composite=Convolution,$domains=({0});
        name=Resolution,Workspace={1},WorkspaceIndex={0};
            (
            name=UserFunction,Formula=
            A*exp( -(x-x0)^2/2/Sigma^2)/(2*3.1415*Sigma^2)^0.5,
            A=1.,x0=0.,Sigma=6.0,  ties=();
                (
                composite=ProductFunction,NumDeriv=false;name=TabulatedFunction,Workspace={2},WorkspaceIndex={0},
                ties=(Scaling=1,Shift=0,XScaling=1);
                name=UserFunction,Formula=
                Sigma*1.4142/12.*exp( -(x)^2/2/Sigma^2)/(2*3.1415*Sigma^2)^0.5
                *((8*((x)/sqrt(2.)/Sigma)^3-12*((x)/sqrt(2.)/Sigma))),
                Sigma=6.0);ties=()
                )
            )"""
    elif yFitIC.fitModel == "GC_C4_C6":
        convolution_template = """
        (composite=Convolution,$domains=({0});
        name=Resolution,Workspace={1},WorkspaceIndex={0};
            (
            name=UserFunction,Formula=
            A*exp( -(x-x0)^2/2/Sigma^2)/(2*3.1415*Sigma^2)^0.5
            *(1+c4/32*(16*((x-x0)/sqrt(2.)/Sigma)^4-48*((x-x0)/sqrt(2.)/Sigma)^2+12)),
            A=1.,x0=0.,Sigma=6.0, c4=0, ties=();
                (
                composite=ProductFunction,NumDeriv=false;name=TabulatedFunction,Workspace={2},WorkspaceIndex={0},
                ties=(Scaling=1,Shift=0,XScaling=1);
                name=UserFunction,Formula=
                Sigma*1.4142/12.*exp( -(x)^2/2/Sigma^2)/(2*3.1415*Sigma^2)^0.5
                *((8*((x)/sqrt(2.)/Sigma)^3-12*((x)/sqrt(2.)/Sigma))),
                Sigma=6.0);ties=()
                )
            )"""
    else:
        raise ValueError(
            f"{yFitIC.fitModel} not supported for Global Mantid fit. Supported options: 'SINGLE_GAUSSIAN',"
            f"'GC_C4_C6'"
        )

    print("\nGlobal fit in the West domain over 8 mixed banks\n")
    widths = []
    for bank in range(8):
        dets = [bank, bank + 8, bank + 16, bank + 24]

        convolvedFunctionsList = []
        ties = ["f0.f1.f1.f1.Sigma=f0.f1.f0.Sigma"]
        datasets = {"InputWorkspace": wsGlobal.name(), "WorkspaceIndex": dets[0]}

        print("Detectors: ", dets)

        counter = 0
        for i in dets:
            print(f"Considering spectrum {wsGlobal.getSpectrumNumbers()[i]}")
            if wsGlobal.spectrumInfo().isMasked(i):
                print(f"Skipping masked spectrum {wsGlobal.getSpectrumNumbers()[i]}")
                continue

            thisIterationFunction = convolution_template.format(
                counter, wsRes.name(), wsQInv.name()
            )
            convolvedFunctionsList.append(thisIterationFunction)

            if counter > 0:
                if yFitIC.fitModel == "SINGLE_GAUSSIAN":
                    ties.append(
                        "f{0}.f1.f0.Sigma= f{0}.f1.f1.f1.Sigma=f0.f1.f0.Sigma".format(
                            counter
                        )
                    )
                else:
                    ties.append(
                        "f{0}.f1.f0.Sigma= f{0}.f1.f1.f1.Sigma=f0.f1.f0.Sigma".format(
                            counter
                        )
                    )
                    ties.append("f{0}.f1.f0.c4=f0.f1.f0.c4".format(counter))
                    ties.append("f{0}.f1.f1.f1.c3=f0.f1.f1.f1.c3".format(counter))

                # Attach datasets
                datasets[f"InputWorkspace_{counter}"] = wsGlobal.name()
                datasets[f"WorkspaceIndex_{counter}"] = i
            counter += 1

        multifit_func = f"composite=MultiDomainFunction; {';'.join(convolvedFunctionsList)}; ties=({','.join(ties)})"
        minimizer_string = (
            f"{minimizer}, AbsError=0.00001, RealError=0.00001, MaxIterations=2000"
        )

        # Unpack dictionary as arguments
        Fit(
            multifit_func,
            Minimizer=minimizer_string,
            Output=wsFirstMass.name() + f"_Joy_Mixed_Banks_Bank_{str(bank)}_fit",
            **datasets,
        )

        # Select ws with fit results
        ws = mtd[
            wsFirstMass.name() + f"_Joy_Mixed_Banks_Bank_{str(bank)}_fit_Parameters"
        ]
        print(f"Bank: {str(bank)} -- sigma={ws.cell(2,1)} +/- {ws.cell(2,2)}")
        widths.append(ws.cell(2, 1))

        DeleteWorkspace(
            wsFirstMass.name()
            + "_Joy_Mixed_Banks_Bank_"
            + str(bank)
            + "_fit_NormalisedCovarianceMatrix"
        )
        DeleteWorkspace(
            wsFirstMass.name()
            + "_Joy_Mixed_Banks_Bank_"
            + str(bank)
            + "_fit_Workspaces"
        )

    print(
        "\nAverage hydrogen standard deviation: ",
        np.mean(widths),
        " +/- ",
        np.std(widths),
    )
    return widths



from mvesuvio.vesuvio_analysis.ICHelpers import (
    buildFinalWSName,
    completeICFromInputs,
    completeBootIC,
    completeYFitIC,
)
from mvesuvio.vesuvio_analysis.bootstrap import runBootstrap
from mvesuvio.vesuvio_analysis.fit_in_yspace import fitInYSpaceProcedure
from mvesuvio.vesuvio_analysis.procedures import (
    runIndependentIterativeProcedure,
    runJointBackAndForwardProcedure,
    runPreProcToEstHRatio,
    createTableWSHRatios,
    isHPresent,
)
from mantid.api import mtd


def runScript(
    userCtr,
    scriptName,
    wsBackIC,
    wsFrontIC,
    bckwdIC,
    fwdIC,
    yFitIC,
    bootIC,
    yes_to_all=False,
):
    # Set extra attributes from user attributes
    completeICFromInputs(fwdIC, scriptName, wsFrontIC)
    completeICFromInputs(bckwdIC, scriptName, wsBackIC)
    completeBootIC(bootIC, bckwdIC, fwdIC, yFitIC)
    completeYFitIC(yFitIC, scriptName)

    checkInputs(userCtr)
    checkInputs(bootIC)
    assert not (
        userCtr.runRoutine & bootIC.runBootstrap
    ), "Main routine and bootstrap both set to run!"

    def runProcedure():
        proc = userCtr.procedure  # Shorthad to make it easier to read

        if proc is None:
            return

        ranPreliminary = False
        if (proc == "BACKWARD") | (proc == "JOINT"):
            if isHPresent(fwdIC.masses) & (bckwdIC.HToMassIdxRatio is None):
                HRatios, massIdxs = runPreProcToEstHRatio(
                    bckwdIC, fwdIC
                )  # Sets H ratio to bckwdIC automatically
                ranPreliminary = True
            assert isHPresent(fwdIC.masses) != (
                bckwdIC.HToMassIdxRatio is None
            ), "When H is not present, HToMassIdxRatio has to be set to None"

        if proc == "BACKWARD":
            res = runIndependentIterativeProcedure(bckwdIC)
        if proc == "FORWARD":
            res = runIndependentIterativeProcedure(fwdIC)
        if proc == "JOINT":
            res = runJointBackAndForwardProcedure(bckwdIC, fwdIC)

        # If preliminary procedure ran, make TableWS with H ratios values
        if ranPreliminary:
            createTableWSHRatios(HRatios, massIdxs)
        return res

    # Names of workspaces to be fitted in y space
    wsNames = []
    ICs = []
    for mode, IC in zip(["BACKWARD", "FORWARD"], [bckwdIC, fwdIC]):
        if (userCtr.fitInYSpace == mode) | (userCtr.fitInYSpace == "JOINT"):
            wsNames.append(buildFinalWSName(scriptName, mode, IC))
            ICs.append(IC)

    # If bootstrap is not None, run bootstrap procedure and finish
    if bootIC.runBootstrap:
        assert (
            (bootIC.procedure == "FORWARD")
            | (bootIC.procedure == "BACKWARD")
            | (bootIC.procedure == "JOINT")
        ), "Invalid Bootstrap procedure."
        return runBootstrap(bckwdIC, fwdIC, bootIC, yFitIC), None

    # Default workflow for procedure + fit in y space
    if userCtr.runRoutine:
        # Check if final ws are loaded:
        wsInMtd = [ws in mtd for ws in wsNames]  # Bool list
        if (len(wsInMtd) > 0) and all(
            wsInMtd
        ):  # When wsName is empty list, loop doesn't run
            for wsName, IC in zip(wsNames, ICs):
                resYFit = fitInYSpaceProcedure(yFitIC, IC, mtd[wsName])
            return None, resYFit  # To match return below.

        checkUserClearWS(yes_to_all)  # Check if user is OK with cleaning all workspaces
        res = runProcedure()

        resYFit = None
        for wsName, IC in zip(wsNames, ICs):
            resYFit = fitInYSpaceProcedure(yFitIC, IC, mtd[wsName])

        return res, resYFit  # Return results used only in tests


def checkUserClearWS(yes_to_all=False):
    """If any workspace is loaded, check if user is sure to start new procedure."""

    if not yes_to_all and len(mtd) != 0:
        userInput = input(
            "This action will clean all current workspaces to start anew. Proceed? (y/n): "
        )
        if (userInput == "y") | (userInput == "Y"):
            pass
        else:
            raise KeyboardInterrupt("Run of procedure canceled.")
    return


def checkInputs(crtIC):
    try:
        if ~crtIC.runRoutine:
            return
    except AttributeError:
        if ~crtIC.runBootstrap:
            return

    for flag in [crtIC.procedure, crtIC.fitInYSpace]:
        assert (
            (flag == "BACKWARD")
            | (flag == "FORWARD")
            | (flag == "JOINT")
            | (flag is None)
        ), "Option not recognized."

    if (crtIC.procedure != "JOINT") & (crtIC.fitInYSpace is not None):
        assert crtIC.procedure == crtIC.fitInYSpace



from .analysis_functions import iterativeFitForDataReduction
from mantid.api import AnalysisDataService
from mantid.simpleapi import CreateEmptyTableWorkspace
import numpy as np


def runIndependentIterativeProcedure(IC, clearWS=True):
    """
    Runs the iterative fitting of NCP, cleaning any previously stored workspaces.
    input: Backward or Forward scattering initial conditions object
    output: Final workspace that was fitted, object with results arrays
    """

    # Clear worksapces before running one of the procedures below
    if clearWS:
        AnalysisDataService.clear()

    return iterativeFitForDataReduction(IC)


def runJointBackAndForwardProcedure(bckwdIC, fwdIC, clearWS=True):
    assert (
        bckwdIC.modeRunning == "BACKWARD"
    ), "Missing backward IC, args usage: (bckwdIC, fwdIC)"
    assert (
        fwdIC.modeRunning == "FORWARD"
    ), "Missing forward IC, args usage: (bckwdIC, fwdIC)"

    # Clear worksapces before running one of the procedures below
    if clearWS:
        AnalysisDataService.clear()

    return runJoint(bckwdIC, fwdIC)


def runPreProcToEstHRatio(bckwdIC, fwdIC):
    """
    Used when H is present and H to first mass ratio is not known.
    Preliminary forward scattering is run to get rough estimate of H to first mass ratio.
    Runs iterative procedure with alternating back and forward scattering.
    """

    assert (
        bckwdIC.runningSampleWS is False
    ), "Preliminary procedure not suitable for Bootstrap."
    fwdIC.runningPreliminary = True

    # Store original no of MS and set MS iterations to zero
    oriMS = []
    for IC in [bckwdIC, fwdIC]:
        oriMS.append(IC.noOfMSIterations)
        IC.noOfMSIterations = 0

    nIter = askUserNoOfIterations()

    HRatios = []  # List to store HRatios
    massIdxs = []
    # Run preliminary forward with a good guess for the widths of non-H masses
    wsFinal, fwdScatResults = iterativeFitForDataReduction(fwdIC)
    for i in range(int(nIter)):  # Loop until convergence is achieved
        AnalysisDataService.clear()  # Clears all Workspaces

        # Update H ratio
        massIdx, HRatio = calculateHToMassIdxRatio(fwdScatResults)
        bckwdIC.HToMassIdxRatio = HRatio
        bckwdIC.massIdx = massIdx
        HRatios.append(HRatio)
        massIdxs.append(massIdx)

        wsFinal, bckwdScatResults, fwdScatResults = runJoint(bckwdIC, fwdIC)

    print(f"\nIdxs of masses for H ratio for each iteration: \n{massIdxs}")
    print(f"\nCorresponding H ratios: \n{HRatios}")

    fwdIC.runningPreliminary = (
        False  # Change to default since end of preliminary procedure
    )

    # Set original number of MS iterations
    for IC, ori in zip([bckwdIC, fwdIC], oriMS):
        IC.noOfMSIterations = ori

    # Update the H ratio with the best estimate, chages bckwdIC outside function
    massIdx, HRatio = calculateHToMassIdxRatio(fwdScatResults)
    bckwdIC.HToMassIdxRatio = HRatio
    bckwdIC.massIdx = massIdx
    HRatios.append(HRatio)
    massIdxs.append(massIdx)

    return HRatios, massIdxs


def createTableWSHRatios(HRatios, massIdxs):
    tableWS = CreateEmptyTableWorkspace(
        OutputWorkspace="H_Ratios_From_Preliminary_Procedure"
    )
    tableWS.setTitle("H Ratios and Idxs at each iteration")
    tableWS.addColumn(type="int", name="iter")
    tableWS.addColumn(type="float", name="H Ratio")
    tableWS.addColumn(type="int", name="Mass Idx")
    for i, (hr, hi) in enumerate(zip(HRatios, massIdxs)):
        tableWS.addRow([i, hr, hi])
    return


def askUserNoOfIterations():
    print("\nH was detected but HToMassIdxRatio was not provided.")
    print(
        "\nSugested preliminary procedure:\n\nrun_forward\nfor n:\n    estimate_HToMassIdxRatio\n    run_backward\n"
        "    run_forward"
    )
    userInput = input(
        "\n\nDo you wish to run preliminary procedure to estimate HToMassIdxRatio? (y/n)"
    )
    if not ((userInput == "y") or (userInput == "Y")):
        raise KeyboardInterrupt("Preliminary procedure interrupted.")

    nIter = int(input("\nHow many iterations do you wish to run? n="))
    return nIter


def calculateHToMassIdxRatio(fwdScatResults):
    """
    Calculate H ratio to mass with highest peak.
    Returns idx of mass and corresponding H ratio.
    """
    fwdMeanIntensityRatios = fwdScatResults.all_mean_intensities[-1]

    # To find idx of mass in backward scattering, take out first mass H
    fwdIntensitiesNoH = fwdMeanIntensityRatios[1:]

    massIdx = np.argmax(
        fwdIntensitiesNoH
    )  # Idex of forward inensities, which include H
    assert (
        fwdIntensitiesNoH[massIdx] != 0
    ), "Cannot estimate H intensity since maximum peak from backscattering is zero."

    HRatio = fwdMeanIntensityRatios[0] / fwdIntensitiesNoH[massIdx]

    return massIdx, HRatio


def runJoint(bckwdIC, fwdIC):
    wsFinal, bckwdScatResults = iterativeFitForDataReduction(bckwdIC)
    setInitFwdParsFromBackResults(bckwdScatResults, bckwdIC, fwdIC)
    wsFinal, fwdScatResults = iterativeFitForDataReduction(fwdIC)
    return wsFinal, bckwdScatResults, fwdScatResults


def setInitFwdParsFromBackResults(bckwdScatResults, bckwdIC, fwdIC):
    """
    Used to pass mean widths and intensities from back scattering onto intial conditions of forward scattering.
    Checks if H is present and adjust the passing accordingly:
    If H present, use HToMassIdxRatio to recalculate intensities and fix only non-H widths.
    If H not present, widths and intensities are directly mapped and all widhts except first are fixed.
    """

    # Get widts and intensity ratios from backscattering results
    backMeanWidths = bckwdScatResults.all_mean_widths[-1]
    backMeanIntensityRatios = bckwdScatResults.all_mean_intensities[-1]

    if isHPresent(fwdIC.masses):
        assert len(backMeanWidths) == fwdIC.noOfMasses - 1, (
            "H Mass present, no of masses in frontneeds to be bigger" "than back by 1."
        )

        # Use H ratio to calculate intensity ratios
        HIntensity = bckwdIC.HToMassIdxRatio * backMeanIntensityRatios[bckwdIC.massIdx]
        # Add H intensity in the first idx
        initialFwdIntensityRatios = np.append([HIntensity], backMeanIntensityRatios)
        # Normalize intensities
        initialFwdIntensityRatios /= np.sum(initialFwdIntensityRatios)

        # Set calculated intensity ratios to forward scattering
        fwdIC.initPars[0::3] = initialFwdIntensityRatios
        # Set forward widths from backscattering
        fwdIC.initPars[4::3] = backMeanWidths
        # Fix all widths except for H, i.e. the first one
        fwdIC.bounds[4::3] = backMeanWidths[:, np.newaxis] * np.ones((1, 2))

    else:  # H mass not present anywhere
        assert len(backMeanWidths) == fwdIC.noOfMasses, (
            "H Mass not present, no of masses needs to be the same for"
            "front and back scattering."
        )

        # Set widths and intensity ratios
        fwdIC.initPars[1::3] = backMeanWidths
        fwdIC.initPars[0::3] = backMeanIntensityRatios

        if len(backMeanWidths) > 1:  # In the case of single mass, width is not fixed
            # Fix all widhts except first
            fwdIC.bounds[4::3] = backMeanWidths[1:][:, np.newaxis] * np.ones((1, 2))

    print(
        "\nChanged initial conditions of forward scattering according to mean widhts and intensity ratios from "
        "backscattering.\n"
    )
    return


def isHPresent(masses) -> bool:
    Hmask = np.abs(masses - 1) / 1 < 0.1  # H mass whithin 10% of 1 au

    if np.any(Hmask):  # H present
        print("\nH mass detected.\n")
        assert (
            len(Hmask) > 1
        ), "When H is only mass present, run independent forward procedure, not joint."
        assert Hmask[0], "H mass needs to be the first mass in masses and initPars."
        assert sum(Hmask) == 1, "More than one mass very close to H were detected."
        return True
    else:
        return False



from mvesuvio.vesuvio_analysis.run_script import runScript
import unittest
import numpy as np
import numpy.testing as nptest
from pathlib import Path
from mvesuvio.system_tests.tests_IC import (
    scriptName,
    wsBackIC,
    wsFrontIC,
    bckwdIC,
    fwdIC,
    yFitIC,
)


class BootstrapInitialConditions:
    runBootstrap = True

    procedure = "JOINT"
    fitInYSpace = "FORWARD"

    bootstrapType = "BOOT_RESIDUALS"
    nSamples = 3
    skipMSIterations = False
    runningTest = True
    userConfirmation = False


class UserScriptControls:
    runRoutine = False
    procedure = "FORWARD"
    fitInYSpace = None
    # bootstrap = "JOINT"


class TestJointBootstrap(unittest.TestCase):
    @classmethod
    def _run_analysis(cls):
        np.random.seed(1)  # Set seed so that tests match everytime
        bootIC = BootstrapInitialConditions
        userCtr = UserScriptControls

        # Change yFItIC to default settings, running tests for yfit before hand changes this
        yFitIC.fitModel = "SINGLE_GAUSSIAN"
        yFitIC.symmetrisationFlag = True

        bootRes, noneRes = runScript(
            userCtr, scriptName, wsBackIC, wsFrontIC, bckwdIC, fwdIC, yFitIC, bootIC, True
        )

        # TODO: Figure out why doing the two tests simultaneously fails the testing
        # Probably because running bootstrap alters the initial conditions of forward scattering
        # Test Joint procedure

        cls._bootBackSamples = bootRes["bckwdScat"].bootSamples
        cls._bootFrontSamples = bootRes["fwdScat"].bootSamples
        cls._bootYFitSamples = bootRes["fwdYFit"].bootSamples

        # bootJointResults = bootRes

        # bootSamples = []
        # for bootRes in bootJointResults:
        #     bootSamples.append(bootRes.bootSamples)

        # bootBackSamples, bootFrontSamples, bootYFitSamples = bootSamples

    @classmethod
    def _load_benchmark_results(cls):
        testPath = Path(__file__).absolute().parent
        cls._oriJointBack = np.load(str(testPath / "stored_boot_back.npz"))[
            "boot_samples"
        ]
        cls._oriJointFront = np.load(str(testPath / "stored_boot_front.npz"))[
            "boot_samples"
        ]
        cls._oriJointYFit = np.load(str(testPath / "stored_boot_yfit.npz"))[
            "boot_samples"
        ]

    @classmethod
    def setUpClass(cls):
        cls._run_analysis()
        cls._load_benchmark_results()

    def xtestBack(self):
        nptest.assert_array_almost_equal(self._bootBackSamples, self._oriJointBack)

    def xtestFront(self):
        nptest.assert_array_almost_equal(self._bootFrontSamples, self._oriJointFront)

    def xtestYFit(self):
        nptest.assert_array_almost_equal(self._bootYFitSamples, self._oriJointYFit)


# # Test Single procedure
# class TestIndependentBootstrap(unittest.TestCase):
#     def _run_analysis(self):
#         np.random.seed(1)  # Set seed so that tests match everytime
#         bootIC = BootstrapInitialConditions
#         userCtr = UserScriptControls
#
#         # Change yFItIC to default settings, running tests for yfit before hand changes this
#         yFitIC.fitModel = "SINGLE_GAUSSIAN"
#         yFitIC.symmetrisationFlag = True
#
#         bootSingleResults = runIndependentBootstrap(bckwdIC, bootIC, yFitIC)
#
#         self._bootSingleBackSamples = bootSingleResults[0].bootSamples
#         self._bootSingleYFitSamples = bootSingleResults[1].bootSamples
#
#     def _load_benchmark_results(self):
#         testPath = Path(__file__).absolute().parent
#         self._oriBack = np.load(testPath / "stored_single_boot_back.npz")["boot_samples"]
#         self._oriYFit = np.load(testPath / "stored_single_boot_back_yfit.npz")["boot_vals"]
#
#     def setUp(self):
#         self._run_analysis()
#         self._load_benchmark_results()
#
#     def testBack(self):
#         nptest.assert_array_almost_equal(self._bootSingleBackSamples, self._oriBack)
#
#     def testYFit(self):
#         nptest.assert_array_almost_equal(self._bootSingleYFitSamples, self._oriYFit)
#

if __name__ == "__main__":
    unittest.main()



"""
Vesuvio
=============

Vesuvio is an instrument that performs Neuton Compton Scattering, based at ISIS, RAL, UK. This code processes raw output data to determine
nuclear kinetic energies and moment distributions.
"""

version_info = (0, 0, 0)
__version__ = ".".join(map(str, version_info))
__project_url__ = "https://github.com/mantidproject/vesuvio"

from mvesuvio.scripts import main

class ArgInputs:
    def __init__(self, command):
        self.__command = command

    @property
    def command(self):
        return self.__command

class ConfigArgInputs(ArgInputs):
    def __init__(self, set_cache, set_experiment, set_ipfolder):
        super().__init__("config")
        self.__set_cache = set_cache
        self.__set_experiment = set_experiment
        self.__set_ipfolder = set_ipfolder
    
    @property
    def set_cache(self):
        return self.__set_cache
    
    @property
    def set_experiment(self):
        return self.__set_experiment

    @property
    def set_ipfolder(self):
        return self.__set_ipfolder


class RunArgInputs(ArgInputs):
    def __init__(self, yes):
        super().__init__("run")
        self.__yes = yes

    @property
    def yes(self):
        return self.__yes


def set_config(cache_directory="", experiment_id="", ip_folder=""):
    config_args = ConfigArgInputs(cache_directory, experiment_id, ip_folder)
    main(config_args)
    
def run(yes_to_all=False):
    run_args = RunArgInputs(yes_to_all)
    main(run_args)



from xml.dom import NotFoundErr
from mvesuvio.vesuvio_analysis.analysis_functions import (
    calculateMeansAndStds,
    filterWidthsAndIntensities,
)
from mvesuvio.vesuvio_analysis.ICHelpers import setBootstrapDirs
from mvesuvio.vesuvio_analysis.fit_in_yspace import selectModelAndPars
import numpy as np
import matplotlib.pyplot as plt
from scipy import stats


def runAnalysisOfStoredBootstrap(bckwdIC, fwdIC, yFitIC, bootIC, analysisIC, userCtr):
    if not (analysisIC.runAnalysis):
        return

    setBootstrapDirs(
        bckwdIC, fwdIC, bootIC, yFitIC
    )  # Same function used to store data, to check below if dirs exist

    for IC in [bckwdIC, fwdIC]:
        if not (IC.bootSavePath.is_file()):
            print("Bootstrap data files not found, unable to run analysis!")
            print(f"{IC.bootSavePath.name}")
            continue  # If main results are not present, assume ysapce results are also missing

        checkLogMatch(IC, isYFitFile=False)

        bootParsRaw, parentParsRaw, nSamples, corrResiduals = readBootData(
            IC.bootSavePath
        )
        checkResiduals(corrResiduals)
        checkBootSamplesVSParent(bootParsRaw, parentParsRaw, IC)  # Prints comparison

        bootPars = (
            bootParsRaw.copy()
        )  # By default do not filter means, copy to avoid accidental changes
        if analysisIC.filterAvg:
            bootPars = filteredBootMeans(bootParsRaw.copy(), IC)
            try:
                print("\nCompare filtered parameters with parent:\n")
                checkBootSamplesVSParent(
                    bootPars, parentParsRaw, IC
                )  # Prints comparison
            except AssertionError:
                print("\nUnable to calculate new means of filtered parameters.\n")

        # Plots histograms of all spectra for a given width or intensity
        plotRawWidthsAndIntensities(analysisIC, IC, bootPars, parentParsRaw)

        # Calculate bootstrap histograms for mean widths and intensities
        meanWidths, meanIntensities = calculateMeanWidthsIntensities(
            bootPars, IC, nSamples
        )

        # If filer is on, check that it matches original procedure
        checkMeansProcedure(analysisIC, IC, meanWidths, meanIntensities, bootParsRaw)

        plotMeanWidthsAndIntensities(
            analysisIC, IC, meanWidths, meanIntensities, parentParsRaw
        )
        plotMeansEvolution(analysisIC, meanWidths, meanIntensities)
        plot2DHistsWidthsAndIntensities(analysisIC, meanWidths, meanIntensities)

        if not (IC.bootYFitSavePath.is_file()):
            print(
                "Bootstrap data file for y-space fit not found, unable to run analysis!"
            )
            print(f"{IC.bootYFitSavePath.name}")
            continue

        checkLogMatch(IC, isYFitFile=True)

        minuitFitVals = readYFitData(IC.bootYFitSavePath, yFitIC)

        plotMeansEvolutionYFit(analysisIC, minuitFitVals)
        plotYFitHists(analysisIC, yFitIC, minuitFitVals)
        plot2DHistsYFit(analysisIC, minuitFitVals)


def checkLogMatch(IC, isYFitFile):
    """Checks if currently selected data file matches stored logs."""
    currentLog = IC.bootYFitSavePathLog if isYFitFile else IC.bootSavePathLog
    currName = currentLog.split(" : ")[0]
    # Check sample present in log file
    with open(IC.logFilePath, "r") as logFile:
        for line in logFile:
            # Check if name of file is present
            if line.split(" : ")[0] == currName:
                if line.strip("\n") == currentLog:
                    return
                raise NotFoundErr(
                    currName + " found but corresponding log does not match."
                )
        raise NotFoundErr(currName + " not found in logs file")


def readBootData(dataPath):
    bootData = np.load(dataPath)

    # Select fitting parameters of ncp
    # Discard first column of spectrum number
    # Discard last two columns of number of iterations and chi2
    bootParsRaw = bootData["boot_samples"][:, :, 1:-2]
    parentParsRaw = bootData["parent_result"][:, 1:-2]

    nSamples = len(bootParsRaw)
    try:
        corrResiduals = bootData["corr_residuals"]
    except KeyError:
        corrResiduals = np.array([np.nan])
        print("\nCorrelation of coefficients not found!\n")

    failMask = np.all(np.isnan(bootParsRaw), axis=(1, 2))
    assert failMask.shape == (
        len(bootParsRaw),
    ), f"Wrong shape of masking: {failMask.shape} != {bootParsRaw.shape} "
    if np.sum(failMask) > 0:
        print(f"\nNo of failed samples: {np.sum(failMask)}")
        print("\nUsing only good replicas ...\n")
        bootParsRaw = bootParsRaw[~failMask]
        nSamples = np.sum(~failMask)

    maskedIdxs = np.where(np.all(np.isnan(bootParsRaw), axis=(0, 2)))
    print(f"Masked idxs with nans found: {maskedIdxs}")
    print(f"\nData files found:\n{dataPath.name}")
    print(f"\nNumber of samples in the file: {nSamples}")
    assert ~np.all(
        bootParsRaw[-1] == parentParsRaw
    ), "Error in Jackknife due to last column."
    return bootParsRaw, parentParsRaw, nSamples, corrResiduals


def readYFitData(dataPath, yFitIC):
    """Resulting output has shape(no of pars, no of samples)"""

    fitIdx = 0  # 0 to select Minuit, 1 to select LM

    bootYFitData = np.load(dataPath)
    try:
        bootYFitVals = bootYFitData["boot_samples"]
    except KeyError:
        bootYFitVals = bootYFitData["boot_vals"]  # To account for some previous samples
    minFitVals = bootYFitVals[
        :, fitIdx, :-1
    ].T  # Select Minuit values and Discard last value chi2

    failMask = np.all(np.isnan(minFitVals), axis=0)
    if np.sum(failMask) > 0:
        print(f"\nNo of failed samples: {np.sum(failMask)}")
        print("\nUsing only good replicas ...\n")
        minFitVals = minFitVals[:, ~failMask]

    try:
        parentPopt = bootYFitData["parent_popt"][fitIdx]
        parentPerr = bootYFitData["parent_perr"][fitIdx]
        printYFitParentPars(yFitIC, parentPopt, parentPerr)  # TODO: Test this function
    except KeyError:
        pass
    return minFitVals


def checkResiduals(corrRes):
    if np.all(np.isnan(corrRes)):
        return

    corrCoef = corrRes[:, 0]
    nCorrelated = np.sum(corrCoef > 0.5)
    print(f"\nNumber of spectra with pearson r > 0.5: {nCorrelated}")
    return


def checkBootSamplesVSParent(bestPars, parentPars, IC):
    """
    For an unbiased estimator, the mean of the bootstrap samples will converge to
    the mean of the experimental sample (here called parent).
    """

    bootWidths = bestPars[:, :, 1::3]
    bootIntensities = bestPars[:, :, 0::3]

    meanBootWidths = np.mean(bootWidths, axis=0)
    meanBootIntensities = np.mean(bootIntensities, axis=0)

    avgWidths, stdWidths, avgInt, stdInt = calculateMeansAndStds(
        meanBootWidths.T, meanBootIntensities.T, IC
    )

    parentWidths = parentPars[:, 1::3]
    parentIntensities = parentPars[:, 0::3]

    avgWidthsP, stdWidthsP, avgIntP, stdIntP = calculateMeansAndStds(
        parentWidths.T, parentIntensities.T, IC
    )

    print("\nComparing Bootstrap means with parent means:\n")
    printResults(avgWidths, stdWidths, "Boot Widths")
    printResults(avgWidthsP, stdWidthsP, "Parent Widths")
    printResults(avgInt, stdInt, "Boot Intensities")
    printResults(avgIntP, stdIntP, "Parent Intensities")


def printResults(arrM, arrE, mode):
    print(f"\n{mode}:\n")
    for i, (m, e) in enumerate(zip(arrM, arrE)):
        print(f"{mode} {i}: {m:>6.3f} \u00B1 {e:<6.3f}")


def filteredBootMeans(
    bestPars, IC
):  # Pass IC just to check flag for preliminary procedure
    """Use same filtering function used on original procedure"""

    # Extract Widths and Intensities from bootstrap samples
    bootWidths = bestPars[:, :, 1::3]
    bootIntensities = bestPars[:, :, 0::3]

    # Perform the filter
    for i, (widths, intensities) in enumerate(zip(bootWidths, bootIntensities)):
        filteredWidths, filteredIntensities = filterWidthsAndIntensities(
            widths.T, intensities.T, IC
        )

        bootWidths[i] = filteredWidths.T
        bootIntensities[i] = filteredIntensities.T

    # Convert back to format of bootstrap samples
    filteredBestPars = bestPars.copy()
    filteredBestPars[:, :, 1::3] = bootWidths
    filteredBestPars[:, :, 0::3] = bootIntensities
    return filteredBestPars


def plotRawWidthsAndIntensities(analysisIC, IC, bootPars, parentPars):
    """
    Plots histograms of each width and intensity seperatly.
    Plots histogram of means over spectra for each width or intensity.
    """

    if not (analysisIC.plotRawWidthsIntensities):
        return

    parentWidths, parentIntensities = extractParentMeans(parentPars, IC)
    noOfMasses = len(parentWidths)

    fig, axs = plt.subplots(2, noOfMasses)

    for axIdx, startIdx, kind, parentMeans in zip(
        [0, 1], [1, 0], ["Width", "Intensity"], [parentWidths, parentIntensities]
    ):
        for i, j in enumerate(range(startIdx, 3 * noOfMasses, 3)):
            axs[axIdx, i].set_title(f"{kind} {i}")
            idxSamples = selectRawSamplesPerIdx(bootPars, j)
            plotHists(axs[axIdx, i], idxSamples, disableCI=True, disableLeg=True)
            axs[axIdx, i].axvline(
                parentMeans[i], 0.75, 0.97, color="b", ls="-", alpha=0.4
            )

    plt.show()
    return


def extractParentMeans(parentPars, IC):
    """Uses original treatment of widths and intensities to calculate parent means."""
    # Modify so that the filtering doesn't happen by default
    meanWp, meanIp = calcMeansWithOriginalProc(parentPars[np.newaxis, :, :], IC)
    meanWp = meanWp.flatten()
    meanIp = meanIp.flatten()
    return meanWp, meanIp


def calcMeansWithOriginalProc(bestPars, IC):
    """Performs the means and std on each bootstrap sample according to original procedure"""

    bootWidths = bestPars[:, :, 1::3]
    bootIntensities = bestPars[:, :, 0::3]

    bootMeanW = np.zeros((len(bootWidths[0, 0, :]), len(bootWidths)))
    # bootStdW = np.zeros(bootMeanW.shape)
    bootMeanI = np.zeros(bootMeanW.shape)
    # bootStdI = np.zeros(bootMeanW.shape)

    for j, (widths, intensities) in enumerate(zip(bootWidths, bootIntensities)):
        meanW, stdW, meanI, stdI = calculateMeansAndStds(widths.T, intensities.T, IC)

        bootMeanW[:, j] = meanW  # Interested only in the means
        bootMeanI[:, j] = meanI

    return bootMeanW, bootMeanI


def calculateMeanWidthsIntensities(bootPars, IC, nSamples):
    """
    Calculates means for each Bootstrap sample.
    Returns means with size equal to the number of Boot samples.
    """

    # Calculate bootstrap histograms for mean widths and intensities
    meanWidths = np.zeros((len(IC.masses), nSamples))
    for i, j in enumerate(range(1, 3 * len(IC.masses), 3)):
        idxSamples = selectRawSamplesPerIdx(bootPars, j)
        meanWidths[i, :] = np.nanmean(idxSamples, axis=0)

    meanIntensities = np.zeros((len(IC.masses), nSamples))
    for i, j in enumerate(range(0, 3 * len(IC.masses), 3)):
        idxSamples = selectRawSamplesPerIdx(bootPars, j)
        meanIntensities[i, :] = np.nanmean(idxSamples, axis=0)

    return meanWidths, meanIntensities


def checkMeansProcedure(analysisIC, IC, meanWidths, meanIntensities, bootParsRaw):
    """Checks that filtering and averaging of Bootstrap samples follows the original procedure"""

    if not (
        analysisIC.filterAvg
    ):  # When filtering not present, no comparison available
        return

    else:  # Check that treatment of data matches original
        meanWOri, meanIOri = calcMeansWithOriginalProc(bootParsRaw, IC)
        np.testing.assert_array_almost_equal(meanWOri, meanWidths)
        np.testing.assert_array_almost_equal(meanIOri, meanIntensities)
        return


def plotMeanWidthsAndIntensities(
    analysisIC, IC, meanWidths, meanIntensities, parentParsRaw
):
    """
    Most informative histograms, shows all mean widhts and intensities of Bootstrap samples
    """

    if not (analysisIC.plotMeanWidthsIntensities):
        return

    parentWidths, parentIntensities = extractParentMeans(parentParsRaw, IC)

    print("\n\n Test passed! Mean Widths match!")

    fig, axs = plt.subplots(2, 1)
    axs[0].set_title("Histograms of mean Widths")
    axs[1].set_title("Histograms of mean Intensitiess")

    for ax, means, parentMeans in zip(
        axs.flatten(), [meanWidths, meanIntensities], [parentWidths, parentIntensities]
    ):
        plotHists(ax, means, disableAvg=True, disableCI=True)
        for pMean in parentMeans:
            ax.axvline(pMean, 0.75, 0.97, color="b", ls="-", alpha=0.4)

    plt.show()
    return


def plotMeansEvolution(IC, meanWidths, meanIntensities):
    """Shows how results of Bootstrap change depending on number of bootstrap samples"""

    if not (IC.plotMeansEvolution):
        return

    fig, axs = plt.subplots(2, 2)
    axs[0, 0].set_title("Evolution of mean Widths")
    plotMeansOverNoSamples(axs[0, 0], meanWidths)

    axs[1, 0].set_title("Widths normalized to last value")
    plotMeansOverNoSamples(axs[1, 0], meanWidths, normFlag=True)

    axs[0, 1].set_title("Evolution of mean Intensities")
    plotMeansOverNoSamples(axs[0, 1], meanIntensities)

    axs[1, 1].set_title("Intensities normalized to last value")
    plotMeansOverNoSamples(axs[1, 1], meanIntensities, normFlag=True)

    plt.show()
    return


def plotMeansEvolutionYFit(analysisIC, minuitFitVals):
    if not (analysisIC.plotMeansEvolution):
        return

    fig, ax = plt.subplots(2, 1)
    ax[0].set_title("Evolution of y-space fit parameters")
    plotMeansOverNoSamples(ax[0], minuitFitVals)

    ax[1].set_title("y-space fit parameters normalized to last value")
    plotMeansOverNoSamples(ax[1], minuitFitVals, normFlag=True)

    plt.show()
    return


def plotMeansOverNoSamples(ax, bootMeans, normFlag=False):
    nSamples = len(bootMeans[0])
    assert nSamples >= 10, "To plot evolution of means, need at least 10 samples!"
    noOfPoints = int(nSamples / 10)
    sampleSizes = np.linspace(10, nSamples, noOfPoints).astype(int)

    sampleMeans = np.zeros((len(bootMeans), len(sampleSizes)))
    sampleBounds = np.zeros((len(bootMeans), 2, len(sampleSizes)))

    for i, N in enumerate(sampleSizes):
        subSample = bootMeans[:, :N].copy()

        mean = np.mean(subSample, axis=1)

        bounds = np.percentile(subSample, [16, 68 + 16], axis=1).T  # 1 standard dev
        assert bounds.shape == (len(subSample), 2), f"Wrong shape: {bounds.shape}"
        # errors = bounds - mean[:, np.newaxis]

        sampleMeans[:, i] = mean
        sampleBounds[:, :, i] = bounds

    meansFinal = sampleMeans
    boundsFinal = sampleBounds
    ylabel = "Means and Errors Values"
    if (
        normFlag
    ):  # Rescale and normalize to last value, so all points are converging to zero
        lastValue = sampleMeans[:, -1][:, np.newaxis]
        meansFinal = (
            (sampleMeans - lastValue) / lastValue * 100
        )  # Percent change to last value
        boundsFinal = (
            (sampleBounds - lastValue[:, np.newaxis, :])
            / lastValue[:, np.newaxis, :]
            * 100
        )
        ylabel = "Percent Change [%]"

    for i, (means, errors) in enumerate(zip(meansFinal, boundsFinal)):
        ax.plot(sampleSizes, means, label=f"idx {i}")
        ax.fill_between(sampleSizes, errors[0, :], errors[1, :], alpha=0.2)

    ax.legend()
    ax.set_xlabel("Number of Bootstrap samples")
    ax.set_ylabel(ylabel)


def plot2DHistsWidthsAndIntensities(IC, meanWidths, meanIntensities):
    if not (IC.plot2DHists):
        return

    assert (
        meanWidths.shape == meanIntensities.shape
    ), "Widths and Intensities need to be the same shape."

    plot2DHists(meanWidths, "Widths")
    plot2DHists(meanIntensities, "Intensities")
    return


def plot2DHists(bootSamples, mode):
    """bootSamples has histogram rows for each parameter"""

    plotSize = len(bootSamples)
    fig, axs = plt.subplots(plotSize, plotSize, tight_layout=True)

    for i in range(plotSize):
        for j in range(plotSize):
            # axs[i, j].axis("off")

            if j > i:
                axs[i, j].set_visible(False)

            elif i == j:
                if i > 0:
                    orientation = "horizontal"
                else:
                    orientation = "vertical"

                axs[i, j].hist(bootSamples[i], orientation=orientation)

            else:
                axs[i, j].hist2d(bootSamples[j], bootSamples[i])

            if j == 0:
                axs[i, j].set_ylabel(f"idx {i}")
            else:
                axs[i, j].get_yaxis().set_ticks([])

            if i == plotSize - 1:
                axs[i, j].set_xlabel(f"idx{j}")
            else:
                axs[i, j].get_xaxis().set_ticks([])

            axs[i, j].set_title(
                f"r = {stats.pearsonr(bootSamples[i], bootSamples[j])[0]:.3f}"
            )

    fig.suptitle(f"1D and 2D Histograms of {mode}")

    plt.show()


def printYFitParentPars(yFitIC, parentPopt, parentPerr):
    model, defaultPars, sharedPars = selectModelAndPars(yFitIC.fitModel)
    sig = [p for p in defaultPars]
    sig = ["y0"] + sig  # Add intercept from outside function parameters

    print("\nParent parameters of y-sapce fit:\n")
    for p, m, e in zip(sig, parentPopt, parentPerr):
        print(f"{p:5s}:  {m:8.3f} +/- {e:8.3f}")


def plotYFitHists(analysisIC, yFitIC, yFitHists):
    """Histogram for each parameter of model used for fit in y-space."""

    if not (analysisIC.plotYFitHists):
        return

    # Plot each parameter in an individual histogram
    fig, axs = plt.subplots(
        2, int(np.ceil(len(yFitHists) / 2)), figsize=(12, 7), tight_layout=True
    )

    # To label each histogram, extract signature of function used for the fit
    model, defaultPars, sharedPars = selectModelAndPars(yFitIC.fitModel)
    sig = [
        p for p in defaultPars
    ]  # Assumes defaultPars in same order as signature of the function
    sig = ["y0"] + sig  # Add intercept from outside function parameters

    for i, (ax, hist, par) in enumerate(zip(axs.flatten(), yFitHists, sig)):
        ax.set_title(f"Fit Parameter: {par}")
        plotHists(ax, hist[np.newaxis, :], disableAvg=True)

    # Hide plots not in use:
    for ax in axs.flat:
        if not ax.lines:  # If empty list
            ax.set_visible(False)

    plt.show()
    return


def plot2DHistsYFit(analysisIC, minuitFitVals):
    if not (analysisIC.plot2DHists):
        return

    plot2DHists(minuitFitVals, "Y-space Fit parameters")
    return


def plotHists(ax, samples, disableCI=False, disableLeg=False, disableAvg=False):
    """Plots each row of 2D samples array as a histogram."""

    # ax.set_title(f"Histogram of {title}")
    for i, bootHist in enumerate(samples):
        if np.all(bootHist == 0) or np.all(np.isnan(bootHist)):
            continue

        mean = np.nanmean(bootHist)
        bounds = np.percentile(bootHist, [16, 68 + 16])  # 1 std: 68%, 2 std: 95%
        errors = bounds - mean
        leg = f"Row {i}: {mean:>6.3f} +{errors[1]:.3f} {errors[0]:.3f}"

        ax.hist(bootHist, histtype="step", label=leg, linewidth=1)
        ax.axvline(mean, 0.9, 0.97, color="k", ls="--", alpha=0.4)

        if not (disableCI):
            ax.axvspan(bounds[0], bounds[1], alpha=0.1, color="b")

    if not (disableAvg):
        # Plot average over histograms
        avgHist = np.nanmean(samples, axis=0).flatten()
        ax.hist(avgHist, color="r", histtype="step", linewidth=2)
        ax.axvline(np.nanmean(avgHist), 0.75, 0.97, color="r", ls="--", linewidth=2)

    if not (disableLeg):
        ax.legend(loc="upper center")


def selectRawSamplesPerIdx(bootSamples, idx):
    """
    Returns samples and mean for a single width or intensity at a given index.
    Samples shape: (No of spectra, No of boot samples)
    Samples mean: Mean of spectra (mean of histograms)
    """
    samples = bootSamples[:, :, idx].T
    return samples


# TODO: Make use of this function?
def calcCorrWithScatAngle(samples, IC):
    """Calculate correlation coefficient between histogram means and scattering angle."""

    ipMatrix = np.loadtxt(IC.instrParsPath, dtype=str)[1:].astype(float)

    firstSpec, lastSpec = IC.bootSavePath.name.split("_")[1].split("-")
    allSpec = ipMatrix[:, 0]
    selectedSpec = (allSpec >= int(firstSpec)) & (allSpec >= int(lastSpec))

    thetas = ipMatrix[selectedSpec, 2]  # Scattering angle on third column

    # thetas = ipMatrix[firstIdx : lastIdx, 2]    # Scattering angle on third column
    assert thetas.shape == (len(samples),), f"Wrong shape: {thetas.shape}"

    histMeans = np.nanmean(samples, axis=1)

    # Remove masked spectra:
    nanMask = np.isnan(histMeans)
    histMeans = histMeans[~nanMask]
    thetas = thetas[~nanMask]

    corr = stats.pearsonr(thetas, histMeans)
    return corr



import numpy as np


class LoadVesuvioBackParameters:
    def __init__(self, ipFilesPath):
        self.ipfile = ipFilesPath / "ip2019.par"

    runs = "43066-43076"  # 77K         # The numbers of the runs to be analysed
    empty_runs = (
        "41876-41923"  # 77K         # The numbers of the empty runs to be subtracted
    )
    spectra = "3-134"  # Spectra to be analysed
    mode = "DoubleDifference"
    subEmptyFromRaw = True  # Flag to control wether empty ws gets subtracted from raw
    scaleEmpty = 1  # None or scaling factor
    scaleRaw = 1


class LoadVesuvioFrontParameters:
    def __init__(self, ipFilesPath):
        self.ipfile = ipFilesPath / "ip2018_3.par"

    runs = "43066-43076"  # 100K        # The numbers of the runs to be analysed
    empty_runs = (
        "43868-43911"  # 100K        # The numbers of the empty runs to be subtracted
    )
    spectra = "144-182"  # Spectra to be analysed
    mode = "SingleDifference"
    subEmptyFromRaw = False  # Flag to control wether empty ws gets subtracted from raw
    scaleEmpty = 1  # None or scaling factor
    scaleRaw = 1


class GeneralInitialConditions:
    transmission_guess = 0.8537  # Experimental value from VesuvioTransmission
    multiple_scattering_order, number_of_events = 2, 1.0e5
    vertical_width, horizontal_width, thickness = 0.1, 0.1, 0.001  # Expressed in meters


class BackwardInitialConditions(GeneralInitialConditions):
    def __init__(self, ipFilesPath):
        self.InstrParsPath = ipFilesPath / "ip2018_3.par"

    HToMassIdxRatio = 19.0620008206  # Set to zero or None when H is not present
    massIdx = 0
    masses = np.array([12, 16, 27])
    initPars = np.array([1, 12, 0.0, 1, 12, 0.0, 1, 12.5, 0.0])
    bounds = np.array(
        [
            [0, np.nan],
            [8, 16],
            [-3, 1],
            [0, np.nan],
            [8, 16],
            [-3, 1],
            [0, np.nan],
            [11, 14],
            [-3, 1],
        ]
    )
    constraints = ()
    noOfMSIterations = 3  # 4
    firstSpec = 3  # 3
    lastSpec = 134  # 134
    maskedSpecAllNo = np.array([18, 34, 42, 43, 59, 60, 62, 118, 119, 133])
    MSCorrectionFlag = True
    GammaCorrectionFlag = False
    tofBinning = "275.,1.,420"  # Binning of ToF spectra
    maskTOFRange = None
    runHistData = True
    normVoigt = False


class ForwardInitialConditions(GeneralInitialConditions):
    def __init__(self, ipFilesPath):
        self.InstrParsPath = ipFilesPath / "ip2018_3.par"

    masses = np.array([1.0079, 12, 16, 27])
    initPars = np.array([1, 4.7, 0, 1, 12.71, 0.0, 1, 8.76, 0.0, 1, 13.897, 0.0])
    bounds = np.array(
        [
            [0, np.nan],
            [3, 6],
            [-3, 1],
            [0, np.nan],
            [12.71, 12.71],
            [-3, 1],
            [0, np.nan],
            [8.76, 8.76],
            [-3, 1],
            [0, np.nan],
            [13.897, 13.897],
            [-3, 1],
        ]
    )
    constraints = ()
    noOfMSIterations = 3  # 4
    firstSpec = 144  # 144
    lastSpec = 182  # 182
    MSCorrectionFlag = True
    GammaCorrectionFlag = True
    maskedSpecAllNo = np.array([173, 174, 179])
    tofBinning = "110,1,430"  # Binning of ToF spectra
    maskTOFRange = None
    runHistData = True
    normVoigt = False


class YSpaceFitInitialConditions:
    anything = True


class UserScriptControls:
    runRoutine = True
    procedure = "FORWARD"  # Options: None, "BACKWARD", "FORWARD", "JOINT"
    fitInYSpace = None  # Options: None, "BACKWARD", "FORWARD", "JOINT"


class BootstrapInitialConditions:
    runBootstrap = False



import numpy as np
from pathlib import Path
from mvesuvio.scripts import handle_config

ipFilesPath = Path(handle_config.read_config_var("caching.ipfolder"))
ipFilePath = ipFilesPath / "ip2018_3.par"


class LoadVesuvioBackParameters:
    runs = "43066-43076"  # 77K         # The numbers of the runs to be analysed
    empty_runs = (
        "41876-41923"  # 77K         # The numbers of the empty runs to be subtracted
    )
    spectra = "3-134"  # Spectra to be analysed
    mode = "DoubleDifference"
    ipfile = ipFilesPath / "ip2019.par"

    subEmptyFromRaw = True  # Flag to control wether empty ws gets subtracted from raw
    scaleEmpty = 1
    scaleRaw = 1


class LoadVesuvioFrontParameters:
    runs = "43066-43076"  # 100K        # The numbers of the runs to be analysed
    empty_runs = (
        "43868-43911"  # 100K        # The numbers of the empty runs to be subtracted
    )
    spectra = "144-182"  # Spectra to be analysed
    mode = "SingleDifference"
    ipfile = ipFilesPath / "ip2018_3.par"

    subEmptyFromRaw = False  # Flag to control wether empty ws gets subtracted from raw
    scaleEmpty = 1
    scaleRaw = 1


class GeneralInitialConditions:
    """Used to define initial conditions shared by both Back and Forward scattering"""

    transmission_guess = 0.8537  # Experimental value from VesuvioTransmission
    multiple_scattering_order, number_of_events = 2, 1.0e5
    # Sample slab parameters
    vertical_width, horizontal_width, thickness = 0.1, 0.1, 0.001  # Expressed in meters


class BackwardInitialConditions(GeneralInitialConditions):
    InstrParsPath = ipFilesPath / "ip2018_3.par"

    HToMassIdxRatio = 19.0620008206  # Set to zero or None when H is not present
    massIdx = 0  # Idx of mass to take the ratio with
    # Masses, instrument parameters and initial fitting parameters
    masses = np.array([12, 16, 27])
    # noOfMasses = len(masses)

    initPars = np.array(
        [
            # Intensities, NCP widths, NCP centers
            1,
            12,
            0.0,
            1,
            12,
            0.0,
            1,
            12.5,
            0.0,
        ]
    )
    bounds = np.array(
        [
            [0, np.nan],
            [8, 16],
            [-3, 1],
            [0, np.nan],
            [8, 16],
            [-3, 1],
            [0, np.nan],
            [11, 14],
            [-3, 1],
        ]
    )
    constraints = ()

    noOfMSIterations = 0  # 4
    firstSpec = 3  # 3
    lastSpec = 13  # 134

    maskedSpecAllNo = np.array([18, 34, 42, 43, 59, 60, 62, 118, 119, 133])

    # Boolean Flags to control script
    MSCorrectionFlag = True
    GammaCorrectionFlag = False

    # # Parameters of workspaces in input_ws
    tofBinning = "275.,1.,420"  # Binning of ToF spectra
    maskTOFRange = None

    # Only when running tests, to match original script
    runHistData = True
    normVoigt = False


class ForwardInitialConditions(GeneralInitialConditions):
    InstrParsPath = ipFilePath

    masses = np.array([1.0079, 12, 16, 27])
    # noOfMasses = len(masses)

    initPars = np.array(
        [
            # Intensities, NCP widths, NCP centers
            1,
            4.7,
            0,
            1,
            12.71,
            0.0,
            1,
            8.76,
            0.0,
            1,
            13.897,
            0.0,
        ]
    )
    bounds = np.array(
        [
            [0, np.nan],
            [3, 6],
            [-3, 1],
            [0, np.nan],
            [12.71, 12.71],
            [-3, 1],
            [0, np.nan],
            [8.76, 8.76],
            [-3, 1],
            [0, np.nan],
            [13.897, 13.897],
            [-3, 1],
        ]
    )
    constraints = ()

    noOfMSIterations = 1  # 4
    firstSpec = 164  # 144
    lastSpec = 175  # 182

    # Boolean Flags to control script
    MSCorrectionFlag = True
    GammaCorrectionFlag = True

    maskedSpecAllNo = np.array([173, 174, 179])

    tofBinning = "110,1.,430"  # Binning of ToF spectra
    maskTOFRange = None

    # Only when running tests
    runHistData = True
    normVoigt = False


class YSpaceFitInitialConditions:
    showPlots = False
    symmetrisationFlag = True
    rebinParametersForYSpaceFit = "-20, 0.5, 20"  # Needs to be symetric
    fitModel = "SINGLE_GAUSSIAN"
    runMinos = True
    globalFit = None
    nGlobalFitGroups = 4
    maskTypeProcedure = None


# userControls and bootIC defined in corresponding test scripts


wsFrontIC = LoadVesuvioFrontParameters()
wsBackIC = LoadVesuvioBackParameters()  # THIS WAS SET TO FRONT

fwdIC = ForwardInitialConditions()
bckwdIC = BackwardInitialConditions()
yFitIC = YSpaceFitInitialConditions()

scriptName = "test_expr"



"""Package defining entry points.
"""
import argparse
from os import environ, path
from mvesuvio.scripts import handle_config


def main(manual_args=None):
    args = __setup_and_parse_args() if not manual_args else manual_args
    if args.command == "config":
        __setup_config(args)

    if args.command == "run":
        if not handle_config.config_set():
            __setup_config(None)
        __run_analysis(args.yes)

def __setup_and_parse_args():
    parser = __set_up_parser()
    args = parser.parse_args()
    return args

def __set_up_parser():
    parser = argparse.ArgumentParser(
        description="Package to analyse Vesuvio instrument data"
    )
    subparsers = parser.add_subparsers(dest="command", required=True)
    config_parser = subparsers.add_parser("config", help="set mvesuvio configuration")
    config_parser.add_argument(
        "--set-cache", "-c", help="set the cache directory", default="", type=str
    )
    config_parser.add_argument(
        "--set-ipfolder",
        "-p",
        help="set the intrument parameters directory",
        default="",
        type=str,
    )
    config_parser.add_argument(
        "--set-experiment",
        "-e",
        help="set the current experiment",
        default="",
        type=str,
    )

    run_parser = subparsers.add_parser("run", help="run mvesuvio analysis")
    run_parser.add_argument(
        "--yes", "-y", help="Say yes to all input prompts", action="store_true"
    )
    return parser


def __setup_config(args):
    config_dir = handle_config.VESUVIO_CONFIG_PATH
    handle_config.setup_config_dir(config_dir)
    ipfolder_dir = handle_config.VESUVIO_IPFOLDER_PATH

    if handle_config.config_set():
        cache_dir = (
            handle_config.read_config_var("caching.location")
            if not args or not args.set_cache
            else args.set_cache
        )
        experiment = (
            handle_config.read_config_var("caching.experiment")
            if not args or not args.set_experiment
            else args.set_experiment
        )
        ipfolder_dir = (
            handle_config.read_config_var("caching.ipfolder")
            if not args or not args.set_ipfolder
            else args.set_ipfolder
        )
    else:
        cache_dir = config_dir if not args or not args.set_cache else args.set_cache
        experiment = (
            "default" if not args or not args.set_experiment else args.set_experiment
        )
        ipfolder_dir = (
            ipfolder_dir if not args or not args.set_ipfolder else args.set_ipfolder
        )

        handle_config.setup_default_ipfile_dir()

    handle_config.set_config_vars(
        {
            "caching.location": cache_dir,
            "caching.experiment": experiment,
            "caching.ipfolder": ipfolder_dir,
        }
    )
    handle_config.setup_expr_dir(cache_dir, experiment)
    handle_config.check_dir_exists("IP folder", ipfolder_dir)


def __run_analysis(yes_to_all):
    environ["MANTIDPROPERTIES"] = path.join(
        handle_config.VESUVIO_CONFIG_PATH, "Mantid.user.properties"
    )
    from mvesuvio import analysis_runner

    analysis_runner.run(yes_to_all)


if __name__ == "__main__":
    main()



from mvesuvio.vesuvio_analysis.fit_in_yspace import fitInYSpaceProcedure
from mvesuvio.vesuvio_analysis.procedures import (
    runJointBackAndForwardProcedure,
    runIndependentIterativeProcedure,
)
from mvesuvio.vesuvio_analysis.ICHelpers import (
    buildFinalWSName,
    noOfHistsFromTOFBinning,
)
from mantid.api import AnalysisDataService, mtd
from mantid.simpleapi import CloneWorkspace, SaveNexus, Load, SumSpectra
from scipy import stats
import numpy as np
from pathlib import Path
import time
import matplotlib.pyplot as plt

plt.style.use("ggplot")
currentPath = Path(__file__).parent.absolute()


def runBootstrap(bckwdIC, fwdIC, bootIC, yFitIC):
    checkValidInput(bootIC)

    checkOutputDirExists(
        bckwdIC, fwdIC, bootIC
    )  # Checks to see if those directories exits already
    askUserConfirmation(bckwdIC, fwdIC, bootIC)
    AnalysisDataService.clear()

    if bootIC.bootstrapType == "JACKKNIFE":
        return JackknifeProcedure(bckwdIC, fwdIC, bootIC, yFitIC)

    return bootstrapProcedure(bckwdIC, fwdIC, bootIC, yFitIC)


def checkValidInput(bootIC):
    boot = bootIC.bootstrapType
    assert (
        (boot == "JACKKNIFE") | (boot == "BOOT_GAUSS_ERRS") | (boot == "BOOT_RESIDUALS")
    ), "bootstrapType not recognized. Options: 'JACKKNIFE', 'BOOT_GAUSS_ERRS', 'BOOT_RESIDUALS'"


def checkOutputDirExists(bckwdIC, fwdIC, bootIC):
    if bootIC.runningTest:
        return

    proc = bootIC.procedure
    if (proc == "BACKWARD") | (proc == "JOINT"):
        checkOutDirIC(bckwdIC)
    if (proc == "FORWARD") | (proc == "JOINT"):
        checkOutDirIC(fwdIC)
    return


def checkOutDirIC(IC):
    if IC.bootSavePath.is_file() or IC.bootYFitSavePath.is_file():
        print(
            f"\nOutput data files were detected:"
            f"\n{IC.bootSavePath.name}\n{IC.bootYFitSavePath.name}"
            f"\nAborting Run of Bootstrap to prevent overwriting data."
            f"\nTo avoid this issue you can change the number of samples to run."
        )
        raise ValueError("Output data directories already exist. Aborted Bootstrap.")
    return


def JackknifeProcedure(bckwdIC, fwdIC, bootIC, yFitIC):
    assert bootIC.procedure is not None

    proc = bootIC.procedure
    if (proc == "FORWARD") | (proc == "BACKWARD"):
        return bootstrapProcedure(bckwdIC, fwdIC, bootIC, yFitIC)

    elif proc == "JOINT":  # Do the Jackknife procedure separately
        # Run original procedure to change fwdIC from running backward
        runOriginalBeforeBootstrap(bckwdIC, fwdIC, bootIC, yFitIC)

        bootIC.procedure = "BACKWARD"
        bootIC.fitInYSpace = "BACKWARD"
        bckwdJackRes = bootstrapProcedure(bckwdIC, fwdIC, bootIC, yFitIC)

        bootIC.procedure = "FORWARD"
        bootIC.fitInYSpace = "FORWARD"
        fwdJackRes = bootstrapProcedure(bckwdIC, fwdIC, bootIC, yFitIC)

        return {**bckwdJackRes, **fwdJackRes}  # For consistency
    else:
        raise ValueError("Bootstrap procedure not recognized.")


def bootstrapProcedure(bckwdIC, fwdIC, bootIC, yFitIC):
    """
    Main algorithm for the Bootstrap.
    Allows for Jackknife or Bootstrap depending on bool flag set in bootIC.
    Chooses fast or slow (correct) version of bootstrap depending on flag set in bootIC.
    Performs either independent or joint procedure depending of len(inputIC).
    """
    if bootIC.bootstrapType == "JACKKNIFE":
        assert (
            bootIC.procedure != "JOINT"
        ), "'JOINT' mode should not have reached Jackknife here."

    AnalysisDataService.clear()

    parentResults, parentWSnNCPs = runOriginalBeforeBootstrap(
        bckwdIC, fwdIC, bootIC, yFitIC
    )
    corrCoefs = autoCorrResiduals(parentWSnNCPs)

    nSamples = chooseNSamples(bootIC, parentWSnNCPs)

    bootResults = initializeResults(parentResults, nSamples, corrCoefs)
    saveBootstrapLogs(bootResults, bckwdIC, fwdIC)
    parentWSNCPSavePaths = convertWSToSavePaths(parentWSnNCPs)

    iStart, iEnd = chooseLoopRange(bootIC, nSamples)

    # Form each bootstrap workspace and run ncp fit with MS corrections
    for i in range(iStart, iEnd):
        AnalysisDataService.clear()
        plt.close(
            "all"
        )  # Not sure if previous step clears plt figures, so introduced this step to be safe

        try:
            sampleInputWS, parentWS = createSampleWS(
                parentWSNCPSavePaths, i, bootIC
            )  # Creates ith sample
        except JackMaskCol:
            continue  # If Jackknife column already masked, skip to next column

        formSampleIC(bckwdIC, fwdIC, bootIC, sampleInputWS, parentWS)
        try:
            iterResults = runMainProcedure(
                bckwdIC, fwdIC, bootIC, yFitIC
            )  # Conversion to YSpace with masked column
        except AssertionError:
            continue  # If the procedure fails, skip to next iteration

        storeBootIter(bootResults, i, iterResults)  # Stores results for each iteration
        saveBootstrapResults(bootResults, bckwdIC, fwdIC)
    return bootResults


def askUserConfirmation(bckwdIC, fwdIC, bootIC):
    """Estimates running time for all samples and asks the user to confirm the run."""

    if not (bootIC.userConfirmation):  # Skip user confirmation
        return

    tDict = storeRunnningTime(
        fwdIC, bckwdIC, bootIC
    )  # Run times file path stores in bootIC

    proc = bootIC.procedure
    runTime = 0
    if (proc == "BACKWARD") | (proc == "JOINT"):
        runTime += calcRunTime(bckwdIC, tDict["tBackNoMS"], tDict["tBackPerMS"], bootIC)

    if (proc == "FORWARD") | (proc == "JOINT"):
        runTime += calcRunTime(fwdIC, tDict["tFowNoMS"], tDict["tFowPerMS"], bootIC)

    userInput = input(
        f"\n\nEstimated time for Bootstrap procedure: {runTime/60:.1f} hours.\nProceed? (y/n): "
    )
    if (userInput == "y") or (userInput == "Y"):
        return
    else:
        raise KeyboardInterrupt("Bootstrap procedure interrupted.")


def storeRunnningTime(fwdIC, bckwdIC, bootIC):
    """Used to write run times to txt file."""

    savePath = bootIC.runTimesPath

    if not (savePath.is_file()):
        with open(savePath, "w") as txtFile:
            txtFile.write(
                "This file stores run times to estimate Bootstrap total run time."
            )
            txtFile.write("\nTime in minutes.\n\n")

    resDict = {}
    with open(savePath, "r") as txtFile:
        for line in txtFile:
            if line[0] == "{":  # If line contains dictionary
                resDict = eval(line)

    if len(resDict) < 4:
        # ans = input("Did not find necessary information to estimate runtime. Will run a short routine to store an
        # estimate. Please wait until this is finished. Press any key to continue.")
        resDict = buildRunTimes(fwdIC, bckwdIC)

        with open(savePath, "a") as txtFile:
            print(resDict, file=txtFile)

    return resDict


def buildRunTimes(fwdIC, bckwdIC):
    resDict = {}
    for IC, mode in zip([bckwdIC, fwdIC], ["Back", "Fow"]):
        oriMS = IC.noOfMSIterations
        for NIter, key in zip([0, 1], ["NoMS", "PerMS"]):
            IC.noOfMSIterations = NIter
            t0 = time.time()
            runIndependentIterativeProcedure(IC)
            t1 = time.time()
            resDict["t" + mode + key] = (t1 - t0) / 60
        # Restore starting value
        IC.noOfMSIterations = oriMS

        # Correct times of only MS by subtacting time spend on fitting ncps
        resDict["t" + mode + "PerMS"] -= 2 * resDict["t" + mode + "NoMS"]

    return resDict


def calcRunTime(IC, tNoMS, tPerMS, bootIC):
    if bootIC.skipMSIterations:
        timePerSample = tNoMS
    else:
        timePerSample = tNoMS + (IC.noOfMSIterations) * (tNoMS + tPerMS)

    nSamples = bootIC.nSamples
    if bootIC.bootstrapType == "JACKKNIFE":
        nSamples = 3 if bootIC.runningTest else noOfHistsFromTOFBinning(IC)

    return nSamples * timePerSample


def chooseLoopRange(bootIC, nSamples):
    iStart = 0
    iEnd = nSamples
    if bootIC.bootstrapType == "JACKKNIFE" and bootIC.runningTest:
        iStart = int(nSamples / 2)
        iEnd = iStart + 3
    return iStart, iEnd


def runOriginalBeforeBootstrap(bckwdIC, fwdIC, bootIC, yFitIC):
    """Runs unaltered procedure to store parent results and select parent ws"""

    setICsToDefault(bckwdIC, fwdIC, yFitIC)
    parentResults = runMainProcedure(bckwdIC, fwdIC, bootIC, yFitIC)
    parentWSnNCPs = selectParentWorkspaces(bckwdIC, fwdIC, bootIC)

    return parentResults, parentWSnNCPs


def chooseNSamples(bootIC, parentWSnNCPs: dict):
    """
    Returns number of samples to run.
    If Jackknife is running, no of samples is the number of bins in the workspace."""

    nSamples = bootIC.nSamples
    if bootIC.bootstrapType == "JACKKNIFE":
        assert (
            len(parentWSnNCPs) == 2
        ), "Running Jackknife, supports only one IC at a time."
        if bootIC.procedure == "FORWARD":
            key = "fwdNCP"
        elif bootIC.procedure == "BACKWARD":
            key = "bckwdNCP"

        nSamples = parentWSnNCPs[
            key
        ].blocksize()  # Number of cols from ncp workspace, accounts for missing last col or not
    return nSamples


def setICsToDefault(bckwdIC, fwdIC, yFitIC):
    """Disables some features of yspace fit, makes sure the default"""

    # Disable Minos
    if yFitIC.runMinos:
        yFitIC.runMinos = False

    # Disable global fit
    if yFitIC.globalFit:
        yFitIC.globalFit = False

    # Don't show plots
    if yFitIC.showPlots:
        yFitIC.showPlots = False

    if bckwdIC.runningSampleWS:
        bckwdIC.runningSampleWS = False
    if fwdIC.runningSampleWS:
        fwdIC.runningSampleWS = False
    return


def runMainProcedure(bckwdIC, fwdIC, bootIC, yFitIC):
    """Decides main procedure to run based on the initial conditions offered as inputs."""

    resultsDict = {}

    if (bootIC.procedure == "FORWARD") | (bootIC.procedure == "BACKWARD"):
        for mode, IC, key in zip(
            ["FORWARD", "BACKWARD"], [fwdIC, bckwdIC], ["fwd", "bckwd"]
        ):
            if bootIC.procedure == mode:
                wsFinal, bckwdScatRes = runIndependentIterativeProcedure(
                    IC, clearWS=False
                )
                resultsDict[key + "Scat"] = bckwdScatRes

                if bootIC.bootstrapType == "JACKKNIFE":
                    yFitIC.maskTypeProcedure = (
                        "NAN"  # Enable NAN averaging in y-space fit
                    )

                bckwdYFitRes = fitInYSpaceProcedure(yFitIC, IC, wsFinal)
                resultsDict[key + "YFit"] = bckwdYFitRes

    elif bootIC.procedure == "JOINT":
        ws, bckwdScatRes, fwdScatRes = runJointBackAndForwardProcedure(
            bckwdIC, fwdIC, clearWS=False
        )
        resultsDict["bckwdScat"] = bckwdScatRes
        resultsDict["fwdScat"] = fwdScatRes

        for mode, IC, key in zip(
            ["FORWARD", "BACKWARD"], [fwdIC, bckwdIC], ["fwd", "bckwd"]
        ):
            if (bootIC.fitInYSpace == mode) | (bootIC.fitInYSpace == "JOINT"):
                wsName = buildFinalWSName(IC.scriptName, mode, IC)
                fwdYFitRes = fitInYSpaceProcedure(yFitIC, IC, mtd[wsName])
                resultsDict[key + "YFit"] = fwdYFitRes
    else:
        raise ValueError("Bootstrap procedure not recognized.")

    return resultsDict


def selectParentWorkspaces(bckwdIC, fwdIC, bootIC):
    """
    Selects parent workspace from which the Bootstrap replicas will be created.
    If fast mode, the parent ws is the final ws after MS corrections.
    """
    parentWSnNCPsDict = {}

    for mode, IC, key in zip(
        ["FORWARD", "BACKWARD"], [fwdIC, bckwdIC], ["fwd", "bckwd"]
    ):
        if (bootIC.procedure == mode) | (bootIC.procedure == "JOINT"):
            wsIter = (
                str(IC.noOfMSIterations) if bootIC.skipMSIterations else "0"
            )  # In case of skipping MS, select very last corrected ws

            parentWS = mtd[IC.name + wsIter]
            parentNCP = mtd[parentWS.name() + "_TOF_Fitted_Profiles"]

            parentWSnNCPsDict[key + "WS"] = parentWS
            parentWSnNCPsDict[key + "NCP"] = parentNCP

    return parentWSnNCPsDict


def autoCorrResiduals(parentWSnNCP: dict):
    """
    Calculates the self-correlation of residuals for each spectrum.
    """
    corrCoefs = {}
    for mode in ["bckwd", "fwd"]:
        try:  # Look for workspaces in dictionary, skip if not present
            parentWS = parentWSnNCP[mode + "WS"]
            parentNCP = parentWSnNCP[mode + "NCP"]
        except KeyError:
            continue

        totNcp = parentNCP.extractY()[:, :]
        dataY = parentWS.extractY()[:, : totNcp.shape[1]]  # Missing last column or not
        residuals = dataY - totNcp

        lag = 1  # For lag-plot of self-correlation
        corr = np.zeros((len(residuals), 2))
        for i, rowRes in enumerate(residuals):
            corr[i] = stats.pearsonr(rowRes[:-lag], rowRes[lag:])

        corrCoefs[mode + "Scat"] = corr
    return corrCoefs


def initializeResults(parentResults: dict, nSamples, corrCoefs):
    """
    Initializes a list with objects to store output data.
    [BootBackResults, BootFrontResults, BootYSpaceResults]
    """
    bootResultObjs = {}

    for key in ["fwd", "bckwd"]:
        if key + "Scat" in parentResults:
            bootResultObjs[key + "Scat"] = BootScattResults(
                parentResults[key + "Scat"], nSamples, corrCoefs[key + "Scat"]
            )

        if key + "YFit" in parentResults:
            bootResultObjs[key + "YFit"] = BootYFitResults(
                parentResults[key + "YFit"], nSamples
            )
    return bootResultObjs


class BootScattResults:
    def __init__(self, parentResults, nSamples, corr):
        self.parentResult = parentResults.all_spec_best_par_chi_nit[-1]
        self.bootSamples = np.full((nSamples, *self.parentResult.shape), np.nan)
        self.corrResiduals = corr

    def storeBootIterResults(self, j, bootResult):
        self.bootSamples[j] = bootResult.all_spec_best_par_chi_nit[-1]

    def saveResults(self, IC):
        np.savez(
            IC.bootSavePath,
            boot_samples=self.bootSamples,
            parent_result=self.parentResult,
            corr_residuals=self.corrResiduals,
        )

    def saveLog(self, IC):
        with open(IC.logFilePath, "a") as logFile:
            logFile.write("\n" + IC.bootSavePathLog)


class BootYFitResults:
    def __init__(self, parentResults, nSamples):
        self.parentPopt = parentResults.popt
        self.parentPerr = parentResults.perr
        self.bootSamples = np.full((nSamples, *self.parentPopt.shape), np.nan)

    def storeBootIterResults(self, j, bootResult):
        self.bootSamples[j] = bootResult.popt

    def saveResults(self, IC):
        np.savez(
            IC.bootYFitSavePath,
            boot_samples=self.bootSamples,
            parent_popt=self.parentPopt,
            parent_perr=self.parentPerr,
        )

    def saveLog(self, IC):
        with open(IC.logFilePath, "a") as logFile:
            logFile.write("\n" + IC.bootYFitSavePathLog)


def storeBootIter(bootResultObjs: dict, j: int, bootIterResults: dict):
    for key in bootResultObjs:
        bootResultObjs[key].storeBootIterResults(j, bootIterResults[key])
    return


def saveBootstrapResults(bootResultObjs: dict, bckwdIC, fwdIC):
    for key, IC in zip(["bckwd", "fwd"], [bckwdIC, fwdIC]):
        for res in ["Scat", "YFit"]:
            if key + res in bootResultObjs:
                bootResultObjs[key + res].saveResults(IC)
    return


def saveBootstrapLogs(bootResultObjs: dict, bckwdIC, fwdIC):
    for key, IC in zip(["bckwd", "fwd"], [bckwdIC, fwdIC]):
        for res in ["Scat", "YFit"]:
            if key + res in bootResultObjs:
                bootResultObjs[key + res].saveLog(IC)
    return


def convertWSToSavePaths(parentWSnNCPs: dict):
    savePaths = {}
    for key in parentWSnNCPs:
        savePaths[key] = saveWorkspacesLocally(parentWSnNCPs[key])
    return savePaths


def saveWorkspacesLocally(ws):
    keys = ws.name().split("_")
    saveName = "Parent"

    if "FORWARD" in keys:
        saveName += "_Front"
    elif "BACKWARD" in keys:
        saveName += "_Back"

    if "Profiles" in keys:
        saveName += "_NCP"

    saveName += ".nxs"
    savePath = currentPath / "bootstrap_ws" / saveName
    SaveNexus(ws, str(savePath))
    return savePath


def createSampleWS(parentWSNCPSavePaths: dict, j: int, bootIC):
    boot = bootIC.bootstrapType
    if boot == "JACKKNIFE":
        return createJackknifeWS(parentWSNCPSavePaths, j)
    elif boot == "BOOT_RESIDUALS":
        return createBootstrapWS(parentWSNCPSavePaths)
    elif boot == "BOOT_GAUSS_ERRS":
        return createBootstrapWS(parentWSNCPSavePaths, drawGauss=True)


def createBootstrapWS(parentWSNCPSavePaths: dict, drawGauss=False):
    """
    Creates bootstrap ws replica.
    Inputs: Experimental (parent) workspace and corresponding NCP total fit
    """

    bootInputWS = {}
    parentInputWS = {}
    for key in ["bckwd", "fwd"]:
        try:
            parentWSPath = parentWSNCPSavePaths[key + "WS"]
            totNcpWSPath = parentWSNCPSavePaths[key + "NCP"]
        except KeyError:
            continue

        parentWS, totNcpWS = loadWorkspacesFromPath(parentWSPath, totNcpWSPath)

        totNcp = totNcpWS.extractY()[:, :]
        dataY = parentWS.extractY()[:, : totNcp.shape[1]]  # Missing last col or not
        dataE = parentWS.extractE()[:, : totNcp.shape[1]]

        # Filter out masked columns
        maskCols = np.all(dataY == 0, axis=0)
        dataY, totNcp, dataE = (
            dataY[:, ~maskCols],
            totNcp[:, ~maskCols],
            dataE[:, ~maskCols],
        )

        # Draw DataY from Gaussian distribution
        if drawGauss:
            bootDataY = np.random.normal(dataY, dataE)  # Mean at dataY, width dataE

        else:  # Default, resample residuals
            residuals = dataY - totNcp
            bootRes = bootstrapResidualsSample(residuals)
            bootDataY = totNcp + bootRes

        # Add masked columns as in parent workspace
        fullBootDataY = np.zeros((len(bootDataY), len(maskCols)))
        fullBootDataY[:, ~maskCols] = bootDataY  # Set non-masked values

        # Pass dataY onto workspace
        wsBoot = CloneWorkspace(
            parentWS, OutputWorkspace=parentWS.name() + "_Bootstrap"
        )
        for i, row in enumerate(fullBootDataY):
            wsBoot.dataY(i)[: len(row)] = row  # Last column will be ignored or not
            if drawGauss:
                wsBoot.dataE(i)[:] = np.zeros(wsBoot.readE(i).size)

        assert ~np.all(
            wsBoot.extractY() == parentWS.extractY()
        ), "Bootstrap data not being correctly passed onto ws."

        bootInputWS[key + "WS"] = wsBoot
        parentInputWS[key + "WS"] = parentWS
        parentInputWS[key + "NCP"] = totNcpWS
    return bootInputWS, parentInputWS


def bootstrapResidualsSample(residuals):
    """Randomly choose points from residuals of each spectra (same statistical weigth)"""

    bootRes = np.zeros(residuals.shape)
    for i, res in enumerate(residuals):
        rowIdxs = np.random.randint(0, len(res), len(res))  # [low, high)
        bootRes[i] = res[rowIdxs]
    return bootRes


def createJackknifeWS(parentWSNCPSavePaths: list, j: int):
    """
    Creates jackknife ws replicas.
    Inputs: Experimental (parent) workspace and corresponding NCP total fit
    """

    jackInputWS = {}
    parentInputWS = {}
    # Jackknife does not have 'JOINT' option
    # Careful with this step if in future Jackknife allows for 'JOINT' internally
    assert (
        len(parentWSNCPSavePaths) == 2
    ), "Jackknife can only allow either forward or backward at a time."
    for key in ["bckwd", "fwd"]:  # Only one iteration is selected at a time
        try:
            parentWSPath = parentWSNCPSavePaths[key + "WS"]
            totNcpWSPath = parentWSNCPSavePaths[key + "NCP"]
        except KeyError:
            continue

        parentWS, totNcpWS = loadWorkspacesFromPath(parentWSPath, totNcpWSPath)

        dataY = parentWS.extractY()

        jackDataY = dataY.copy()

        # Skip Jackknife procedure on columns that are already masked
        if np.all(jackDataY[:, j] == 0):
            raise JackMaskCol

        jackDataY[:, j] = 0  # Masks j collumn with zeros
        # DataE is not masked intentionally, to preserve errors that are used in the normalization of averaged NaN profile

        wsJack = CloneWorkspace(
            parentWS, OutputWorkspace=parentWS.name() + "_Jackknife"
        )
        for i, yRow in enumerate(jackDataY):
            wsJack.dataY(i)[:] = yRow  # Last column will be ignored in ncp fit anyway

        assert np.all(
            wsJack.extractY() == jackDataY
        ), "Bootstrap data not being correctly passed onto ws."

        jackInputWS[key + "WS"] = wsJack
        parentInputWS[key + "WS"] = parentWS
        parentInputWS[key + "NCP"] = totNcpWS
    return jackInputWS, parentInputWS


class JackMaskCol(Exception):
    """
    Custom exception used only to flag and skip a Jackknife iteration
    for a column that is already masked.
    """

    pass


def loadWorkspacesFromPath(*savePaths):
    wsList = []
    for path in savePaths:
        saveName = path.name.split(".")[0]
        ws = Load(str(path), OutputWorkspace=saveName)
        SumSpectra(ws, OutputWorkspace=ws.name() + "_Sum")
        wsList.append(ws)

    return wsList


def formSampleIC(bckwdIC, fwdIC, bootIC, sampleInputWS: dict, parentWS: dict):
    """Adds atributes to initial conditions to start procedure with sample ws."""

    for mode, IC, key in zip(
        ["FORWARD", "BACKWARD"], [fwdIC, bckwdIC], ["fwd", "bckwd"]
    ):
        if (bootIC.procedure == mode) | (bootIC.procedure == "JOINT"):
            IC.runningSampleWS = True

            if bootIC.skipMSIterations:
                IC.noOfMSIterations = 0

            IC.sampleWS = sampleInputWS[key + "WS"]
            IC.parentWS = parentWS[key + "WS"]
